"""The API for marshalling higher-order Python objects into simple, encodable Python objects."""

from __future__ import annotations

import typing as tp

from typelib import ctx, graph, serdes
from typelib.marshals import routines
from typelib.py import compat, inspection, refs

T = tp.TypeVar("T")

__all__ = (
    "marshal",
    "marshaller",
    "DelayedMarshaller",
)


def marshal(
    value: tp.Any, *, t: type[T] | refs.ForwardRef | str | None = None
) -> serdes.MarshalledValueT:
    """Marshal `value` from `typ` into [`MarshalledValueT`][typelib.serdes.MarshalledValueT].

    Args:
        value: The value to reduce to a simple, encode-able type.
        t:
            The type to use for building the marshaller (optional).
            If not provided, we'll default to the type of the input value.
    """
    typ = value.__class__ if t is None else t
    routine: routines.AbstractMarshaller[T] = marshaller(typ)
    unmarshalled = routine(value)
    return unmarshalled


@compat.cache
def marshaller(
    t: type[T] | refs.ForwardRef | compat.TypeAliasType | str,
) -> routines.AbstractMarshaller[T]:
    """Get a marshaller routine for a given type.

    Args:
        t:
            The type annotation to generate a marshaller for. Can be a type, type alias,
            [`typing.ForwardRef`][], or string reference.
    """
    nodes = graph.static_order(t)
    context: dict[type | graph.TypeNode, routines.AbstractMarshaller] = (
        ctx.TypeContext()
    )
    if not nodes:
        return routines.NoOpMarshaller(t=t, context=context, var=None)  # type: ignore[arg-type]

    # "root" type will always be the final node in the sequence.
    root = nodes[-1]
    for node in nodes:
        context[node] = _get_unmarshaller(node, context=context)

    return context[root]


def _get_unmarshaller(  # type: ignore[return]
    node: graph.TypeNode,
    context: routines.ContextT,
) -> routines.AbstractMarshaller[T]:
    if node in context:
        return context[node]

    for check, unmarshaller_cls in _HANDLERS.items():
        if check(node.type):
            return unmarshaller_cls(node.type, context=context, var=node.var)

    return routines.StructuredTypeMarshaller(node.type, context=context, var=node.var)


class DelayedMarshaller(routines.AbstractMarshaller[T]):
    """Delayed proxy for a given type's marshaller, used when we encounter a [`typing.ForwardRef`][].

    Notes:
        This allows us to delay the resolution of the given type reference until
        call-time, enabling support for cyclic and recursive types.
    """

    def __init__(
        self, t: type[T], context: routines.ContextT, *, var: str | None = None
    ):
        super().__init__(t, context, var=var)
        self._resolved: routines.AbstractMarshaller[T] | None = None

    @property
    def resolved(self) -> routines.AbstractMarshaller[T]:
        """The resolved marshaller."""
        if self._resolved is None:
            self._resolved = marshaller(self.t)
            for attr in self._resolved.__slots__:
                setattr(self, attr, getattr(self._resolved, attr))
        return self._resolved

    def __call__(self, val: T) -> serdes.MarshalledValueT:
        unmarshalled = self.resolved(val)
        return unmarshalled


# Order is IMPORTANT! This is a FIFO queue.
_HANDLERS: tp.Mapping[
    tp.Callable[[type[T]], bool], type[routines.AbstractMarshaller]
] = {
    # Short-circuit forward refs
    inspection.isforwardref: DelayedMarshaller,
    inspection.isunresolvable: routines.NoOpMarshaller,
    inspection.isnonetype: routines.NoOpMarshaller,
    # Special handler for Literals
    inspection.isliteral: routines.LiteralMarshaller,
    # Special handler for Unions...
    inspection.isuniontype: routines.UnionMarshaller,
    # Non-intersecting types (order doesn't matter here.
    inspection.isdatetimetype: routines.DateTimeMarshaller,
    inspection.isdatetype: routines.DateMarshaller,
    inspection.istimetype: routines.TimeMarshaller,
    inspection.istimedeltatype: routines.TimeDeltaMarshaller,
    inspection.isuuidtype: routines.UUIDMarshaller,
    inspection.ispatterntype: routines.PatternMarshaller,
    inspection.ispathtype: routines.PathMarshaller,
    inspection.isdecimaltype: routines.DecimalMarshaller,
    inspection.isfractiontype: routines.FractionMarshaller,
    inspection.isintegertype: routines.IntegerMarshaller,
    inspection.isfloattype: routines.FloatMarshaller,
    inspection.isstringtype: routines.StringMarshaller,
    inspection.isbytestype: routines.BytesMarshaller,
    # Psuedo-structured containers, should check before generics.
    inspection.istypeddict: routines.StructuredTypeMarshaller,
    inspection.istypedtuple: routines.StructuredTypeMarshaller,
    inspection.isnamedtuple: routines.StructuredTypeMarshaller,
    inspection.isfixedtupletype: routines.FixedTupleMarshaller,
    (
        lambda t: inspection.issubscriptedgeneric(t) and inspection.ismappingtype(t)
    ): routines.SubscriptedMappingMarshaller,
    (
        lambda t: inspection.issubscriptedgeneric(t) and inspection.isiterabletype(t)
    ): routines.SubscriptedIterableMarshaller,
    # A mapping is a collection so must come before that check.
    inspection.ismappingtype: routines.MappingMarshaller,
    # Generic Iterable handler
    inspection.isiterabletype: routines.IterableMarshaller,
}



"""A simple hashmap for working with types in a contextual manner."""

from __future__ import annotations

import dataclasses

from typelib import graph
from typelib.py import inspection, refs


class TypeContext(dict):
    """A key-value mapping which can map between forward references and real types."""

    def __missing__(self, key: graph.TypeNode | type | refs.ForwardRef):
        """Hook to handle missing type references.

        Allows for sharing lookup results between forward references and real types.

        Args:
            key: The type or reference.
        """
        # Eager wrap in a TypeNode
        if not isinstance(key, graph.TypeNode):
            key = graph.TypeNode(type=key)
            return self[key]

        # If we missed a ForwardRef, we've already tried this, bail out.
        type = key.type
        if isinstance(type, refs.ForwardRef):
            raise KeyError(key)

        ref = refs.forwardref(
            inspection.qualname(type), module=getattr(type, "__module__", None)
        )
        node = dataclasses.replace(key, type=ref)
        return self[node]



"""Utilities for working with [`typing.ForwardRef`][].

This module allows the developer to create and evaluate [`typing.ForwardRef`][] instances
with additional logic to support forwards compatibility.

Examples: Typical Usage
    >>> from typelib.py import refs
    >>> ref = refs.forwardref("str")
    >>> cls = refs.evaluate(ref)
    >>> cls is str
    True
"""

from __future__ import annotations

import functools
import inspect
import sys
import typing

from typelib.py import frames, future

__all__ = ("ForwardRef", "evaluate", "forwardref")

ForwardRef: typing.TypeAlias = typing.ForwardRef


def forwardref(
    ref: str,
    *,
    is_argument: bool = False,
    module: typing.Any | None = None,
    is_class: bool = True,
) -> ForwardRef:
    """Create a [`typing.ForwardRef`][] instance from a `ref` string.

    This wrapper function will attempt to determine the module name ahead of instantiation
    if not provided. This is important when resolving the reference to an actual type.

    Args:
        ref: The type reference string.
        is_argument: Whether the reference string was an argument to a function (default False).
        module: The python module in which the reference string is defined (optional)
        is_class: Whether the reference string is a class (default True).
    """
    module = _resolve_module_name(ref, module)
    if module is not None:
        ref = ref.replace(f"{module}.", "")
    return ForwardRef(
        ref,
        is_argument=is_argument,
        module=module,
        is_class=is_class,
    )


if typing.TYPE_CHECKING:

    def evaluate(
        ref: ForwardRef,
        *,
        globalns: typing.Mapping[str, typing.Any] | None = None,
        localns: typing.Mapping[str, typing.Any] | None = None,
        recursive_guard: set | None = None,
    ) -> typing.Any: ...


else:
    if sys.version_info >= (3, 10):

        def evaluate(
            ref: ForwardRef,
            *,
            globalns: typing.Mapping[str, typing.Any] | None = None,
            localns: typing.Mapping[str, typing.Any] | None = None,
            recursive_guard: set | None = None,
        ) -> typing.Any:
            """Evaluate the [`typing.ForwardRef`][] instance into a proper type.

            Note:
                Most of the time you will not need to provide anything but `ref`.

                In Python 3.9, we will automatically transform "new-style" type hints into
                runtime-compatible type hints:
                  - Union-operator (`str | int` -> `typing.Union[str, int]`)
                  - builtin generics (`dict[str, str]` -> `typing.Dict[str, str]`)

            Args:
                ref: The [`typing.ForwardRef`][] instance to evaluate.
                globalns: A mapping of global variable names to values (optional).
                localns: A mapping of local variable names to values (optional).
                recursive_guard: A set instance to prevent recursion during evaluation (optional).
            """
            if type(ref) is not ForwardRef:
                return ref
            if ref.__forward_evaluated__:
                return ref.__forward_value__

            recursive_guard = recursive_guard or set()
            return ref._evaluate(globalns, localns, recursive_guard=recursive_guard)

    else:

        def evaluate(
            ref: ForwardRef,
            *,
            globalns: typing.Mapping[str, typing.Any] | None = None,
            localns: typing.Mapping[str, typing.Any] | None = None,
            recursive_guard: set | None = None,
        ) -> typing.Any:
            if type(ref) is not ForwardRef:
                return ref
            if ref.__forward_evaluated__:
                return ref.__forward_value__

            arg = future.transform_annotation(ref.__forward_arg__)
            globalns = {
                **globals(),
                **(globalns or {}),
                **vars(typing),
                "typing": typing,
            }
            nref = forwardref(
                arg,
                is_argument=ref.__forward_is_argument__,
                module=ref.__forward_module__,
                is_class=ref.__forward_is_class__,
            )
            recursive_guard = recursive_guard or set()
            return nref._evaluate(globalns, localns, recursive_guard)


@functools.cache
def _resolve_module_name(ref: str, module: str | None) -> str | None:
    if module is not None:
        return module

    # Easy path, use the qualname if it's provided.
    module = ref.split(".", maxsplit=1)[0]
    if module != ref:
        return module
    # Harder path, find the actual object in the stack frame, if possible.
    obj = frames.extract(ref)
    module = getattr(obj, "__module__", None)
    if module:
        return module
    # Tricky path, get the caller and get the module name of the caller.
    caller = frames.getcaller()
    module = getattr(inspect.getmodule(caller), "__name__", None)
    return module



"""Utilities for type translation, serialization, and deserialization.

Examples: Typical Usage
    >>> from typelib import serdes
    >>>
    >>> serdes.load("1")
    1

    >>> import datetime
    >>> from typelib import serdes
    >>>
    >>> serdes.unixtime(datetime.datetime(2020, 1, 1))
    1577854800.0
    >>> serdes.isoformat(datetime.timedelta(hours=1))
    'PT1H'

    >>> import dataclasses
    >>> @dataclasses.dataclass
    ... class Class:
    ...     attr: str
    ...
    >>> instance = Class(attr="value")
    >>> dict(serdes.iteritems(instance))
    {'attr': 'value'}
"""

from __future__ import annotations

import ast
import contextlib
import dataclasses
import datetime
import operator
import typing as t

import pendulum
from more_itertools import peekable

from typelib import constants
from typelib.py import compat, inspection

__all__ = (
    "decode",
    "isoformat",
    "unixtime",
    "dateparse",
    "iteritems",
    "itervalues",
    "get_items_iter",
    "strload",
    "load",
)


@t.overload
def decode(  # type: ignore[overload-overlap]
    val: bytes | bytearray | memoryview, *, encoding: str = constants.DEFAULT_ENCODING
) -> str: ...


@t.overload
def decode(val: _T) -> _T: ...  # type: ignore[overload-overlap]


def decode(val: t.Any, *, encoding: str = constants.DEFAULT_ENCODING) -> t.Any:
    """Decode a bytes-like object into a str.

    Note:
        If a non-bytes-like object is passed, it will be returned unchanged.

    Examples:
        >>> from typelib import serdes
        >>> serdes.decode(b"abc")
        'abc'
        >>> serdes.decode(memoryview(b"abc"))
        'abc'

    Args:
        val: The object to be decoded.
        encoding: The encoding to use when decoding the object (defaults "utf8").
    """
    val = val.tobytes() if isinstance(val, memoryview) else val
    if isinstance(val, (bytes, bytearray)):
        decoded = val.decode(encoding)
        return decoded
    return val


@compat.lru_cache(maxsize=100_000)
def isoformat(dt: datetime.date | datetime.time | datetime.timedelta) -> str:
    """Format any date/time object into an ISO-8601 string.

    Note:
        While the standard library includes `isoformat()` methods for
        [`datetime.date`][], [`datetime.time`][], &
        [`datetime.datetime`][], they do not include a method for serializing
        [`datetime.timedelta`][], even though durations are included in the
        ISO 8601 specification. This function fills that gap.

    Examples:
        >>> import datetime
        >>> from typelib import serdes
        >>> serdes.isoformat(datetime.date(1970, 1, 1))
        '1970-01-01'
        >>> serdes.isoformat(datetime.time())
        '00:00:00'
        >>> serdes.isoformat(datetime.datetime(1970, 1, 1))
        '1970-01-01T00:00:00'
        >>> serdes.isoformat(datetime.timedelta(hours=1))
        'PT1H'
    """
    if isinstance(dt, (datetime.date, datetime.time)):
        return dt.isoformat()
    dur: pendulum.Duration = (
        dt
        if isinstance(dt, pendulum.Duration)
        else pendulum.duration(
            days=dt.days,
            seconds=dt.seconds,
            microseconds=dt.microseconds,
        )
    )
    datepart = "".join(
        f"{p}{s}"
        for p, s in ((dur.years, "Y"), (dur.months, "M"), (dur.remaining_days, "D"))
        if p
    )
    timepart = "".join(
        f"{p}{s}"
        for p, s in (
            (dur.hours, "H"),
            (dur.minutes, "M"),
            (
                f"{dur.remaining_seconds}.{dur.microseconds:06}"
                if dur.microseconds
                else dur.remaining_seconds,
                "S",
            ),
        )
        if p
    )
    period = f"P{datepart}T{timepart}"
    return period


_T = t.TypeVar("_T")


def unixtime(dt: datetime.date | datetime.time | datetime.timedelta) -> float:
    """Convert a date/time object to a unix timestamp.

    Note:
        Time is messy. Here is how we've decided to make this work:

        - `datetime.datetime` instances will preserve the current tzinfo (even if naive).
        - `datetime.time` instances will default to today, preserving the tzinfo (even if naive).
        - `datetime.date` instances will assume UTC.
        - `datetime.timedelta` instances be reflected as total seconds since epoch (January 1, 1970).

        If you find yourself in a situation where this does not work for you, your best
        bet is to stop using tz-naive date/time objects. *It's always best to keep your time
        explicit!*

    Args:
        dt: The object to be converted.

    Examples:
        >>> import datetime
        >>> from typelib import serdes
        >>>
        >>> serdes.unixtime(datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc))
        0.0
        >>> serdes.unixtime(datetime.date(1970, 1, 1))
        0.0
    """
    if isinstance(dt, datetime.timedelta):
        return dt.total_seconds()

    if isinstance(dt, datetime.time):
        dt = datetime.datetime.now(tz=dt.tzinfo).replace(
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            microsecond=dt.microsecond,
        )
    if isinstance(dt, datetime.date) and not isinstance(dt, datetime.datetime):
        dt = datetime.datetime(
            year=dt.year,
            month=dt.month,
            day=dt.day,
            tzinfo=datetime.timezone.utc,
        )

    return dt.timestamp()


DateTimeT = t.TypeVar("DateTimeT", datetime.date, datetime.time, datetime.timedelta)


@compat.lru_cache(maxsize=100_000)
def dateparse(val: str, t: type[DateTimeT]) -> DateTimeT:
    """Parse a date string into a datetime object.

    Examples:
        >>> import datetime
        >>> from typelib import serdes
        >>> serdes.dateparse("1970-01-01",t=datetime.datetime)
        DateTime(1970, 1, 1, 0, 0, 0, tzinfo=Timezone('UTC'))

    Args:
        val: The date string to parse.
        t: The target datetime type.

    Returns:
        The parsed datetime object.

    Raises:
        ValueError:
            If `val` is not a date string or does not resolve to an instance of
            the target datetime type.
    """
    try:
        # When `exact=False`, the only two possibilities are DateTime and Duration.
        parsed: pendulum.DateTime | pendulum.Duration = pendulum.parse(val)  # type: ignore[assignment]
        normalized = _nomalize_dt(val=val, parsed=parsed, td=t)
        return normalized
    except ValueError:
        if val.isdigit() or val.isdecimal():
            return _normalize_number(numval=float(val), td=t)
        raise


def _nomalize_dt(
    *, val: str, parsed: pendulum.DateTime | pendulum.Duration, td: type[DateTimeT]
) -> DateTimeT:
    if isinstance(parsed, pendulum.DateTime):
        if issubclass(td, datetime.time):
            return parsed.time().replace(tzinfo=parsed.tzinfo)
        if issubclass(td, datetime.datetime):
            return parsed
        if issubclass(td, datetime.date):
            return parsed.date()
    if not isinstance(parsed, td):
        raise ValueError(f"Cannot parse {val!r} as {td.__qualname__!r}")
    return parsed


def _normalize_number(*, numval: float, td: type[DateTimeT]) -> DateTimeT:
    # Assume the number value is seconds - same logic as time-since-epoch
    if issubclass(td, datetime.timedelta):
        return datetime.timedelta(seconds=numval)
    # Parse a datetime from the time-since-epoch as indicated by the value.
    dt = datetime.datetime.fromtimestamp(numval, tz=datetime.timezone.utc)
    # Return the datetime if the target type is a datetime
    if issubclass(td, datetime.datetime):
        return dt
    # If the target type is a time object, just return the time.
    if issubclass(td, datetime.time):
        return dt.time().replace(tzinfo=dt.tzinfo)
    # If the target type is a date object, just return the date.
    return dt.date()


def iteritems(val: t.Any) -> t.Iterable[tuple[t.Any, t.Any]]:
    """Iterate over (field, value) pairs for any object.

    Examples:
        >>> import dataclasses
        >>> from typelib import serdes
        >>>
        >>> @dataclasses.dataclass
        ... class Class:
        ...     attr: str
        ...
        >>> instance = Class(attr="value")
        >>> [*serdes.iteritems(instance)]
        [('attr', 'value')]
        >>> [*serdes.iteritems("string")]
        [(0, 's'), (1, 't'), (2, 'r'), (3, 'i'), (4, 'n'), (5, 'g')]
        >>> [*serdes.iteritems(serdes.iteritems(instance))]
        [('attr', 'value')]

    Note:
        If the given item is detected to be an iterable of pairs (e.g., `[('a', 1), ('b', 2)]`),
        we will iterate directly over that.

        Otherwise, we will create an iterator over (field, value) pairs with the following
        strategy:

        - For mappings -> `((key, value), ...)`
        - For structured objects (user-defined classes) -> `((field, value), ...)`
        - For all other iterables -> `((index, value), ...)`.

    Args:
        val: The object to iterate over.
    """
    # If the given value is an iterable, we will create a `peekable` so we can inspect it.
    #   In that case, we want to continue using the peekable, since the act of peeking
    #   is destructive to the original iterable in the event it is an iterator.
    is_pairs, it = _is_iterable_of_pairs(val)
    if is_pairs:
        return iter(it)

    iterate = get_items_iter(val.__class__)
    return iterate(it)


def _is_iterable_of_pairs(val: t.Any) -> tuple[bool, t.Any]:
    cls = val.__class__
    if not inspection.isiterabletype(cls) or inspection.ismappingtype(cls):
        return False, val

    if inspection.issequencetype(cls):
        peek = next(iter(val), ())
        is_pairs = inspection.iscollectiontype(peek.__class__) and len(peek) == 2
        return is_pairs, val

    it = peekable(val)
    peek = it.peek()
    is_pairs = inspection.iscollectiontype(peek.__class__) and len(peek) == 2
    return is_pairs, it


def itervalues(val: t.Any) -> t.Iterator[t.Any]:
    """Iterate over the contained values for any object.

    Examples:
        >>> import dataclasses
        >>> from typelib import serdes
        >>>
        >>> @dataclasses.dataclass
        ... class Class:
        ...     attr: str
        ...
        >>> instance = Class(attr="value")
        >>> [*serdes.itervalues(instance)]
        ['value']

    Args:
        val: The object to iterate over.
    """
    iterate = get_items_iter(val.__class__)
    return (v for k, v in iterate(val))


@compat.cache
def get_items_iter(tp: type) -> t.Callable[[t.Any], t.Iterable[tuple[t.Any, t.Any]]]:
    """Given a type, return a callable which will produce an iterator over (field, value) pairs.

    Examples:
        >>> import dataclasses
        >>> from typelib import serdes
        >>>
        >>> @dataclasses.dataclass
        ... class Class:
        ...     attr: str
        ...
        >>> instance = Class(attr="value")
        >>> iteritems = get_items_iter(Class)
        >>> next(iteritems(instance))
        ('attr', 'value')

    Args:
        tp: The type to create an iterator for.
    """
    ismapping, isnamedtuple, isiterable = (
        inspection.ismappingtype(tp),
        inspection.isnamedtuple(tp),
        inspection.isiterabletype(tp),
    )
    if ismapping:
        return _itemscaller
    if isnamedtuple:
        return _namedtupleitems
    if isiterable:
        return enumerate
    return _make_fields_iterator(tp)


def _namedtupleitems(val: t.NamedTuple) -> t.Iterable[tuple[str, t.Any]]:
    return zip(val._fields, val)


def _make_fields_iterator(
    tp: type,
) -> t.Callable[[t.Any], t.Iterator[tuple[t.Any, t.Any]]]:
    # If we have a dataclass, use the defined public fields.
    if dataclasses.is_dataclass(tp):
        public_attribs = [
            f.name for f in dataclasses.fields(tp) if not f.name.startswith("_")
        ]
    # Otherwise, try using the public type-hints.
    else:
        attribs = inspection.get_type_hints(tp)
        public_attribs = [k for k in attribs if not k.startswith("_")]
    # If that didn't work, look for `__slots__`.
    if not public_attribs and hasattr(tp, "__slots__"):
        public_attribs = [s for s in tp.__slots__ if not s.startswith("_")]
    # If we located all public attributes, create a factory function for iterating over
    #   these fields and fetching the value from an instance.
    if public_attribs:

        def _iterfields(val: t.Any) -> t.Iterator[tuple[str, t.Any]]:
            return ((a, getattr(val, a)) for a in public_attribs)

        return _iterfields

    # Finally, if all else fails, just use `vars` on the instance.
    def _itervars(val: t.Any) -> t.Iterator[tuple[str, t.Any]]:
        return ((k, v) for k, v in vars(val).items() if not k.startswith("_"))

    return _itervars


def load(val: _T) -> PythonValueT | _T:
    """Attempt to decode `val` if it is a text-like object.

    Otherwise, return `val` unchanged.

    Examples:
        >>> from typelib import serdes
        >>> serdes.load(1)
        1
        >>> serdes.load("1")
        1
        >>> serdes.load("1,2")
        (1, 2)
        >>> serdes.load(b'{"a": 1, "b": 2}')
        {'a': 1, 'b': 2}

    Args:
        val: The value to decode.
    """
    return strload(val) if inspection.istexttype(val.__class__) else val  # type: ignore[arg-type]


@compat.lru_cache(maxsize=100_000)
def strload(val: str | bytes | bytearray | memoryview) -> PythonValueT:
    """Attempt to decode a string-like input into a Python value.

    Examples:
        >>> from typelib import serdes
        >>> serdes.strload("1")
        1
        >>> serdes.strload("1,2")
        (1, 2)
        >>> serdes.strload(b'{"a": 1, "b": 2}')
        {'a': 1, 'b': 2}


    Tip:
        This function is memoized and only safe for text-type inputs.

    See Also:
         - [`load`][typelib.serdes.load]

    Args:
        val: The string-like input to be decoded.
    """
    with contextlib.suppress(ValueError):
        return compat.json.loads(val)

    decoded = decode(val)
    with contextlib.suppress(ValueError, TypeError, SyntaxError):
        return ast.literal_eval(decoded)

    return decoded


PythonPrimitiveT: t.TypeAlias = "bool | int | float | str | None"
"""Type alias for serializable, non-container Python types."""
PythonValueT: t.TypeAlias = (
    "PythonPrimitiveT | "
    "dict[PythonPrimitiveT, PythonValueT] | "
    "list[PythonValueT] | "
    "tuple[PythonValueT, ...] | "
    "set[PythonValueT]"
)
"""Type alias for any Python builtin type."""
MarshalledValueT: t.TypeAlias = "PythonPrimitiveT | dict[PythonPrimitiveT, MarshalledValueT] | list[MarshalledValueT]"
"""Type alias for a Python value which is ready for over-the-wire serialization."""


_itemscaller = operator.methodcaller("items")
_valuescaller = operator.methodcaller("values")



"""Support for marshalling Python data structures into primitive equivalents.

Notes:
    "Marshalling" your data structure prepares it to be serialized into binary and sent
    over the wire, but *does not* serialize it. We keep these stages isolated to ensure
    maximum flexibility and simplicity.

    We ensure that your marshalled data is compatible with Python's built-in
    [`json`][] module. This provides maximum compatibility with most serialization
    protocols by limiting the output to simple Python builtin types:

    - [`bool`][]
    - [`int`][]
    - [`float`][]
    - [`str`][]
    - [`None`][]
    - [`list`][]
    - [`dict`][]

Tip:
    You may use this package directly, but we encourage you to work with our
    high-level API provided by the top-level [`typelib`][] module.

Examples: Typical Usage
    >>> import dataclasses
    >>> import decimal
    >>> from typelib import marshals
    >>>
    >>> @dataclasses.dataclass(slots=True, weakref_slot=True, kw_only=True)
    ... class Struct:
    ...     key: str
    ...     number: decimal.Decimal
    ...
    >>>
    >>> data = Struct(key="some-key", number=decimal.Decimal("1.0"))
    >>> marshals.marshal(data)
    {'key': 'some-key', 'number': '1.0'}
    >>> marshaller = marshals.marshaller(Struct)
    >>> marshaller(data)
    {'key': 'some-key', 'number': '1.0'}

See Also:
    * [`marshal`][typelib.marshals.marshal]
    * [`marshaller`][typelib.marshals.marshaller]
    * [`typelib.codec`][]
"""

from typelib.marshals.api import *
from typelib.marshals.routines import *



"""High-performance, Fine-grained runtime type inspections.

Examples: Typical Usage
    >>> from typelib.py import inspection
    >>> inspection.ismappingtype(dict)
    True
    >>> inspection.isfixedtupletype(tuple[int, str])
    True

"""

from __future__ import annotations

import abc
import builtins
import collections
import dataclasses
import datetime
import decimal
import enum
import fractions
import functools
import inspect
import ipaddress
import numbers
import pathlib
import re
import sqlite3
import types
import typing as tp
import uuid
from collections.abc import Callable as abc_Callable
from operator import attrgetter

from typelib import constants
from typelib.py import compat, contrib, refs

__all__ = (
    "BUILTIN_TYPES",
    "args",
    "isabstract",
    "isbuiltininstance",
    "isbuiltintype",
    "isbuiltinsubtype",
    "isbytestype",
    "isclassvartype",
    "iscollectiontype",
    "isdatetype",
    "isdecimaltype",
    "isdescriptor",
    "isenumtype",
    "isfinal",
    "isfixedtupletype",
    "isforwardref",
    "isfromdictclass",
    "isfrozendataclass",
    "isgeneric",
    "ishashable",
    "isiterabletype",
    "isliteral",
    "ismappingtype",
    "isnamedtuple",
    "isoptionaltype",
    "isproperty",
    "issequencetype",
    "issimpleattribute",
    "isstdlibinstance",
    "isstdlibtype",
    "isstdlibsubtype",
    "isstringtype",
    "istexttype",
    "istimetype",
    "istimedeltatype",
    "istupletype",
    "istypeddict",
    "istypedtuple",
    "isuniontype",
    "isunresolvable",
    "isuuidtype",
    "name",
    "normalize_typevar",
    "origin",
    "should_unwrap",
    "qualname",
)


@compat.cache
def origin(annotation: tp.Any) -> tp.Any:
    """Get the highest-order 'origin'-type for a given type definition.

    Tip:
        For the purposes of this library, if we can resolve to a builtin type, we will.

    Examples:
        >>> from typelib.py import inspection
        >>> from typing import Dict, Mapping, NewType, Optional
        >>> origin(Dict)
        <class 'dict'>
        >>> origin(Mapping)
        <class 'dict'>
        >>> Registry = NewType('Registry', Dict)
        >>> origin(Registry)
        <class 'dict'>
        >>> class Foo: ...
        ...
        >>> origin(Foo)
        <class 'typelib.Foo'>
    """
    # Resolve custom NewTypes.
    actual = resolve_supertype(annotation)

    # Unwrap optional/classvar
    if isclassvartype(actual):
        a = args(actual)
        actual = a[0] if a else actual

    actual = tp.get_origin(actual) or actual

    # provide defaults for generics
    if not isbuiltintype(actual):
        actual = _check_generics(actual)

    if iscallable(actual):
        actual = tp.Callable

    return actual


def _check_generics(hint: tp.Any):
    return GENERIC_TYPE_MAP.get(hint, hint)


GENERIC_TYPE_MAP: dict[type, type] = {
    tp.Sequence: list,
    tp.MutableSequence: list,
    collections.abc.Sequence: list,
    collections.abc.MutableSequence: list,
    tp.Collection: list,
    collections.abc.Collection: list,
    tp.Iterable: list,
    collections.abc.Iterable: list,
    tp.AbstractSet: set,
    tp.MutableSet: set,
    collections.abc.Set: set,
    collections.abc.MutableSet: set,
    tp.Mapping: dict,
    tp.MutableMapping: dict,
    collections.abc.Mapping: dict,
    collections.abc.MutableMapping: dict,
    tp.Hashable: str,
    collections.abc.Hashable: str,
}


def args(annotation: tp.Any) -> tp.Tuple[tp.Any, ...]:
    """Get the args supplied to an annotation, normalizing [`typing.TypeVar`][].

    Note:
        TypeVar normalization follows this strategy:

            -> If the TypeVar is bound
            -----> return the bound type
            -> Else If the TypeVar has constraints
            -----> return a Union of the constraints
            -> Else
            -----> return Any

    Examples:
        >>> from typelib.py import inspection
        >>> from typing import Dict, TypeVar, Any
        >>> T = TypeVar("T")
        >>> args(Dict)
        ()
        >>> args(Dict[str, int])
        (<class 'str'>, <class 'int'>)
        >>> args(Dict[str, T])
        (<class 'str'>, typing.Any)
    """
    a = tp.get_args(annotation)
    if not a:
        a = getattr(annotation, "__args__", a)

    return (*_normalize_typevars(*a),)


def _normalize_typevars(*args: tp.Any) -> tp.Iterable:
    yield from (normalize_typevar(t) if type(t) is tp.TypeVar else t for t in args)


@compat.cache
def normalize_typevar(tvar: tp.TypeVar) -> type[tp.Any]:
    """Reduce a TypeVar to a simple type.

    Note:
        TypeVar normalization follows this strategy:

            -> If the TypeVar is bound
            -----> return the bound type
            -> Else If the TypeVar has constraints
            -----> return a Union of the constraints
            -> Else
            -----> return Any
    """
    if tvar.__bound__:
        return tvar.__bound__
    elif tvar.__constraints__:
        return tp.Union[tvar.__constraints__]  # type: ignore[return-value]
    return tp.Any  # type: ignore[return-value]


@compat.cache
def name(obj: tp.Union[type, refs.ForwardRef, tp.Callable]) -> str:
    """Safely retrieve the name of either a standard object or a type annotation.

    Examples:
        >>> from typelib.py import inspection
        >>> from typing import Dict, Any, TypeVar
        >>> T = TypeVar("T")
        >>> name(Dict)
        'Dict'
        >>> name(Dict[str, str])
        'Dict'
        >>> name(Any)
        'Any'
        >>> name(dict)
        'dict'
    """
    strobj = qualname(obj)
    return strobj.rsplit(".")[-1]


@compat.cache
def qualname(obj: tp.Union[type, refs.ForwardRef, tp.Callable]) -> str:
    """Safely retrieve the qualname of either a standard object or a type annotation.

    Examples:
        >>> from typelib.py import inspection
        >>> from typing import Dict, Any, TypeVar
        >>> T = TypeVar("T")
        >>> qualname(Dict)
        'typing.Dict'
        >>> qualname(Dict[str, str])
        'typing.Dict'
        >>> qualname(Any)
        'typing.Any'
        >>> qualname(dict)
        'dict'
    """
    strobj = str(obj)
    if isinstance(obj, refs.ForwardRef):
        strobj = str(obj.__forward_arg__)  # type: ignore[union-attr]
    is_generic = isgeneric(strobj)
    # We got a typing thing.
    if is_generic:
        # If this is a subscripted generic we should clean that up.
        return strobj.split("[", maxsplit=1)[0]
    # Easy-ish path, use name magix
    qname = getattr(obj, "__qualname__", None)
    nm = getattr(obj, "__name__", None)
    if qname is not None:
        return qname.replace("<locals>.", "")
    if nm is not None:  # pragma: no cover
        return nm
    return strobj


@compat.cache
def resolve_supertype(annotation: type[tp.Any] | types.FunctionType) -> tp.Any:
    """Get the highest-order supertype for a NewType.

    Examples:
        >>> from typelib.py import inspection
        >>> from typing import NewType
        >>> UserID = NewType("UserID", int)
        >>> AdminID = NewType("AdminID", UserID)
        >>> resolve_supertype(AdminID)
        <class 'int'>
    """
    while hasattr(annotation, "__supertype__"):
        annotation = annotation.__supertype__  # type: ignore[union-attr]
    return annotation


def signature(obj: tp.Callable[..., tp.Any] | type[tp.Any]) -> inspect.Signature:
    """Get the signature of a type or callable.

    Also supports TypedDict subclasses
    """
    if inspect.isclass(obj) or isgeneric(obj):
        if istypeddict(obj):
            return typed_dict_signature(obj)
        if istupletype(obj) and not isnamedtuple(obj):
            return tuple_signature(obj)
    return inspect.signature(obj)


cached_signature = compat.cache(signature)


def get_type_hints(
    obj: tp.Union[type, tp.Callable], exhaustive: bool = True
) -> dict[str, type[tp.Any]]:
    """Wrapper for [`typing.get_type_hints`][].

    If [`typing.get_type_hints`][] raises `([NameError][], [TypeError][])`, we will
    default to an empty dict.

    Args:
        obj: The object to inspect.
        exhaustive:
            Whether to pull type hints from the signature of the object if
            none can be found via [`typing.get_type_hints`][]. (defaults True)
    """
    try:
        hints = tp.get_type_hints(obj)
    except (NameError, TypeError):
        hints = {}
    # KW_ONLY is a special sentinel to denote kw-only params in a dataclass.
    #  We don't want to do anything with this hint/field. It's not real.
    hints = {f: t for f, t in hints.items() if t is not compat.KW_ONLY}
    if not hints and exhaustive:
        hints = _hints_from_signature(obj)
    return hints


def _hints_from_signature(obj: tp.Union[type, tp.Callable]) -> dict[str, type[tp.Any]]:
    try:
        params: dict[str, inspect.Parameter] = {**signature(obj).parameters}
    except (TypeError, ValueError):  # pragma: no cover
        return {}
    hints = {}
    for name, param in params.items():
        annotation = param.annotation
        if annotation is param.empty:
            annotation = tp.Any
            hints[name] = annotation
            continue
        if annotation.__class__ is str:
            ref = refs.forwardref(
                annotation, is_argument=True, module=getattr(obj, "__module__", None)
            )
            hints[name] = ref
            continue
        hints[name] = annotation  # pragma: no cover
    return hints


cached_type_hints = compat.cache(get_type_hints)


def simple_attributes(t: type) -> tp.Tuple[str, ...]:
    """Extract all public, static data-attributes for a given type."""
    # If slots are defined, this is the best way to locate static attributes.
    if hasattr(t, "__slots__") and t.__slots__:
        return (
            *(
                f
                for f in t.__slots__
                if not f.startswith("_")
                # JIC - check if this is something fancy.
                and not isinstance(getattr(t, f, ...), _DYNAMIC_ATTRIBUTES)
            ),
        )
    # Otherwise we have to guess. This is inherently faulty, as attributes aren't
    #   always defined on a class before instantiation. The alternative is reverse
    #   engineering the constructor... yikes.
    return (
        *(
            x
            for x, y in inspect.getmembers(t, predicate=issimpleattribute)
            if not x.startswith("_") and not isinstance(y, _DYNAMIC_ATTRIBUTES)
        ),
    )


_DYNAMIC_ATTRIBUTES = (contrib.SQLAMetaData, contrib.sqla_registry)
cached_simple_attributes = compat.cache(simple_attributes)


def typed_dict_signature(obj: tp.Callable) -> inspect.Signature:
    """A little faker for getting the "signature" of a [`typing.TypedDict`][].

    Note:
        Technically, these are dicts at runtime, but we are enforcing a static shape,
        so we should be able to declare a matching signature for it.
    """
    hints = cached_type_hints(obj)
    total = getattr(obj, "__total__", True)
    default = inspect.Parameter.empty if total else ...
    return inspect.Signature(
        parameters=tuple(
            inspect.Parameter(
                name=x,
                kind=inspect.Parameter.KEYWORD_ONLY,
                annotation=y,
                default=getattr(obj, x, default),
            )
            for x, y in hints.items()
        )
    )


def tuple_signature(t: type[compat.TupleT]) -> inspect.Signature:
    """A little faker for getting the "signature" of a [`tuple`][].

    Note:
        At runtime, tuples are just tuples, but we can make use of their type hints to
        define a predictable signature.
    """
    a = args(t)
    if not a or a[-1] is ...:
        argt = tp.Any if not a else a[0]
        param = inspect.Parameter(
            name="args", kind=inspect.Parameter.VAR_POSITIONAL, annotation=argt
        )
        sig = inspect.Signature(parameters=(param,))
        return sig
    kind = inspect.Parameter.POSITIONAL_ONLY
    params = tuple(
        inspect.Parameter(name=f"arg{str(i)}", kind=kind, annotation=at)
        for i, at in enumerate(a)
    )
    sig = inspect.Signature(parameters=params)
    return sig


@compat.cache
def safe_get_params(obj: type) -> tp.Mapping[str, inspect.Parameter]:
    """Try to extract the parameters of the given object.

    Return an empty mapping if we encounter an error.
    """
    params: tp.Mapping[str, inspect.Parameter]
    try:
        if ismappingtype(obj) and not istypeddict(obj):
            return {}
        params = cached_signature(obj).parameters
    except (ValueError, TypeError):  # pragma: nocover
        params = {}
    return params


@compat.cache
def isbuiltintype(
    obj: type | types.FunctionType,
) -> compat.TypeIs[type[BuiltIntypeT]]:
    """Check whether the provided object is a builtin-type.

    Note:
        Python stdlib and Python documentation have no "definitive list" of
        builtin-**types**, despite the fact that they are well-known. The closest we have
        is https://docs.python.org/3.7/library/functions.html, which clumps the
        builtin-types with builtin-functions. Despite clumping these types with functions
        in the documentation, these types eval as False when compared to
        [`types.BuiltinFunctionType`][], which is meant to be an alias for the
        builtin-functions listed in the documentation.

        All this to say, here we are with a custom check to determine whether a type is a
        builtin.

    Examples:
        >>> from typing import NewType, Mapping
        >>> isbuiltintype(str)
        True
        >>> isbuiltintype(NewType("MyStr", str))
        True
        >>> class Foo: ...
        ...
        >>> isbuiltintype(Foo)
        False
        >>> isbuiltintype(Mapping)
        False
    """
    return (
        resolve_supertype(obj) in BUILTIN_TYPES
        or resolve_supertype(type(obj)) in BUILTIN_TYPES
    )


@compat.cache
def isstdlibtype(obj: type) -> compat.TypeIs[type[STDLibtypeT]]:
    return (
        resolve_supertype(obj) in STDLIB_TYPES
        or resolve_supertype(type(obj)) in STDLIB_TYPES
    )


@compat.cache
def isbuiltinsubtype(t: type) -> compat.TypeIs[type[BuiltIntypeT]]:
    """Check whether the provided type is a subclass of a builtin-type.

    Examples:
        >>> from typing import NewType, Mapping
        >>> class SuperStr(str): ...
        ...
        >>> isbuiltinsubtype(SuperStr)
        True
        >>> isbuiltinsubtype(NewType("MyStr", SuperStr))
        True
        >>> class Foo: ...
        ...
        >>> isbuiltintype(Foo)
        False
        >>> isbuiltintype(Mapping)
        False
    """
    return issubclass(resolve_supertype(t), BUILTIN_TYPES_TUPLE)


@compat.cache
def isstdlibsubtype(t: type) -> compat.TypeIs[type[STDLibtypeT]]:
    """Test whether the given type is a subclass of a standard-lib type.

    Examples:
        >>> import datetime

        >>> class MyDate(datetime.date): ...
        ...
        >>> isstdlibsubtype(MyDate)
        True
    """
    return _safe_issubclass(resolve_supertype(t), STDLIB_TYPES_TUPLE)


def isbuiltininstance(o: tp.Any) -> compat.TypeIs[BuiltIntypeT]:
    """Test whether an object is an instance of a builtin type.

    Examples:
        >>> isbuiltininstance("")
        True
    """
    return builtins.isinstance(o, BUILTIN_TYPES_TUPLE)


def isstdlibinstance(o: tp.Any) -> compat.TypeIs[STDLibtypeT]:
    """Test whether an object is an instance of a type in the standard-lib."""
    return builtins.isinstance(o, STDLIB_TYPES_TUPLE)


@compat.cache
def isoptionaltype(obj: type[_OT]) -> compat.TypeIs[type[tp.Optional[_OT]]]:
    """Test whether an annotation is [`typing.Optional`][], or can be treated as.

    [`typing.Optional`][] is an alias for `typing.Union[<T>, None]`, so both are
    "optional".

    Examples:
        >>> from typing import Optional, Union, Dict, Literal
        >>> isoptionaltype(Optional[str])
        True
        >>> isoptionaltype(Union[str, None])
        True
        >>> isoptionaltype(Literal["", None])
        True
        >>> isoptionaltype(Dict[str, None])
    False
    """
    args = getattr(obj, "__args__", ())
    tname = name(origin(obj))
    nullarg = next((a for a in args if a in (type(None), None)), ...)
    isoptional = tname == "Optional" or (
        nullarg is not ... and tname in ("Union", "Uniontype", "Literal")
    )
    return isoptional


_OT = tp.TypeVar("_OT")


@compat.cache
def isuniontype(obj: type) -> compat.TypeIs[tp.Union]:
    n = name(origin(obj))
    return n in ("Union", "UnionType")


@compat.cache
def isfinal(obj: type) -> bool:
    """Test whether an annotation is [`typing.Final`][].

    Examples:
        >>> from typing import NewType
        >>> from typelib.py.compat import Final
        >>> isfinal(Final[str])
        True
        >>> isfinal(NewType("Foo", Final[str]))
        True
    """
    return origin(obj) is compat.Final


@compat.cache
def isliteral(obj) -> bool:
    """Test whether an annotation is [`typing.Literal`][].

    Examples:
        >>>
    """
    return origin(obj) is tp.Literal or (
        obj.__class__ is refs.ForwardRef and obj.__forward_arg__.startswith("Literal")
    )


@compat.cache
def isdatetype(
    obj: type,
) -> compat.TypeIs[type[tp.Union[datetime.datetime, datetime.date]]]:
    """Test whether this annotation is a a date/datetime object.

    Examples:
        >>> import datetime
        >>> from typing import NewType
        >>> isdatetype(datetime.datetime)
        True
        >>> isdatetype(datetime.date)
        True
        >>> isdatetype(NewType("Foo", datetime.datetime))
        True
    """
    return builtins.issubclass(origin(obj), datetime.date)


@compat.cache
def isdatetimetype(
    obj: type,
) -> compat.TypeIs[type[tp.Union[datetime.datetime, datetime.date]]]:
    """Test whether this annotation is a a date/datetime object.

    Examples:
        >>> import datetime
        >>> from typing import NewType
        >>> isdatetype(datetime.datetime)
        True
        >>> isdatetype(datetime.date)
        True
        >>> isdatetype(NewType("Foo", datetime.datetime))
        True
    """
    return builtins.issubclass(origin(obj), datetime.datetime)


@compat.cache
def istimetype(obj: type) -> compat.TypeIs[type[datetime.time]]:
    """Test whether this annotation is a a date/datetime object.

    Examples:
        >>> import datetime
        >>> from typing import NewType
        >>> istimetype(datetime.time)
        True
        >>> istimetype(NewType("Foo", datetime.time))
        True
    """
    return builtins.issubclass(origin(obj), datetime.time)


@compat.cache
def istimedeltatype(obj: type) -> compat.TypeIs[type[datetime.timedelta]]:
    """Test whether this annotation is a a date/datetime object.

    Examples:
        >>> import datetime
        >>> from typing import NewType
        >>> istimedeltatype(datetime.timedelta)
        True
        >>> istimedeltatype(NewType("Foo", datetime.timedelta))
        True
    """
    return builtins.issubclass(origin(obj), datetime.timedelta)


@compat.cache
def isdecimaltype(obj: type) -> compat.TypeIs[type[decimal.Decimal]]:
    """Test whether this annotation is a Decimal object.

    Examples:
        >>> import decimal
        >>> from typing import NewType
        >>> isdecimaltype(decimal.Decimal)
        True
        >>> isdecimaltype(NewType("Foo", decimal.Decimal))
        True
    """
    return builtins.issubclass(origin(obj), decimal.Decimal)


@compat.cache
def isfractiontype(obj: type) -> compat.TypeIs[type[fractions.Fraction]]:
    """Test whether this annotation is a Decimal object.

    Examples:
        >>> import fractions
        >>> from typing import NewType
        >>> isdecimaltype(fractions.Fraction)
        True
        >>> isdecimaltype(NewType("Foo", fractions.Fraction))
        True
    """
    return builtins.issubclass(origin(obj), fractions.Fraction)


@compat.cache
def isuuidtype(obj: type) -> compat.TypeIs[type[uuid.UUID]]:
    """Test whether this annotation is a a date/datetime object.

    Examples:
        >>> import uuid
        >>> from typing import NewType
        >>> isuuidtype(uuid.UUID)
        True
        >>> class MyUUID(uuid.UUID): ...
        ...
        >>> isuuidtype(MyUUID)
        True
        >>> isuuidtype(NewType("Foo", uuid.UUID))
        True
    """
    return builtins.issubclass(origin(obj), uuid.UUID)


@compat.cache
def isiterabletype(obj: type) -> compat.TypeIs[type[tp.Iterable]]:
    """Test whether the given type is iterable.

    Examples:
        >>> from typing import Sequence, Collection
        >>> isiterabletype(Sequence[str])
        True
        >>> isiterabletype(Collection)
        True
        >>> isiterabletype(str)
        True
        >>> isiterabletype(tuple)
        True
        >>> isiterabletype(int)
        False
    """
    obj = origin(obj)
    return builtins.issubclass(obj, tp.Iterable)


@compat.cache
def isiteratortype(obj: type) -> compat.TypeIs[type[tp.Iterator]]:
    """Check whether the given object is a subclass of an Iterator.

    Examples:
        >>> def mygen(): yield 1
        ...
        >>> isiteratortype(mygen().__class__)
        True
        >>> isiteratortype(iter([]).__class__)
        True
        >>> isiteratortype(mygen)
        False
        >>> isiteratortype(list)
        False
    """
    obj = origin(obj)
    return builtins.issubclass(obj, tp.Iterator)


@compat.cache
def istupletype(
    obj: tp.Callable[..., tp.Any] | type[tp.Any],
) -> compat.TypeIs[type[tuple]]:
    """Tests whether the given type is a subclass of [`tuple`][].

    Examples:
        >>> from typing import NamedTuple, Tuple
        >>> class MyTup(NamedTuple):
        ...     field: int
        ...
        >>> istupletype(tuple)
        True
        >>> istupletype(Tuple[str])
        True
        >>> istupletype(MyTup)
        True
    """
    obj = origin(obj)
    return obj is tuple or issubclass(obj, tuple)  # type: ignore[arg-type]


@compat.cache
def issequencetype(obj: type) -> compat.TypeIs[type[tp.Collection]]:
    """Test whether this annotation is a subclass of [`typing.Collection`][].

    Includes builtins.

    Examples:
        >>> from typing import Collection, Mapping, NewType, Sequence
        >>> issequencetype(Sequence)
        True
        >>> issequencetype(Mapping[str, str])
        True
        >>> issequencetype(str)
        True
        >>> issequencetype(list)
        True
        >>> issequencetype(NewType("Foo", dict))
        True
        >>> issequencetype(int)
        False
    """
    obj = origin(obj)
    return obj in _COLLECTIONS or builtins.issubclass(obj, tp.Sequence)


@compat.cache
def iscollectiontype(obj: type) -> compat.TypeIs[type[tp.Collection]]:
    """Test whether this annotation is a subclass of [`typing.Collection`][].

    Includes builtins.

    Examples:
        >>> from typing import Collection, Mapping, NewType
        >>> iscollectiontype(Collection)
        True
        >>> iscollectiontype(Mapping[str, str])
        True
        >>> iscollectiontype(str)
        True
        >>> iscollectiontype(list)
        True
        >>> iscollectiontype(NewType("Foo", dict))
        True
        >>> iscollectiontype(int)
        False
    """
    obj = origin(obj)
    return obj in _COLLECTIONS or builtins.issubclass(obj, tp.Collection)


_COLLECTIONS = {list, set, tuple, frozenset, dict, str, bytes}


@compat.cache
def issubscriptedcollectiontype(
    obj: type[tp.Generic[_ArgsT]],  # type: ignore[valid-type]
) -> compat.TypeIs[type[tp.Collection[_ArgsT]]]:
    """Test whether this annotation is a collection type and is subscripted.

    Examples:
        >>> from typing import Collection, Mapping, NewType
        >>> issubscriptedcollectiontype(Collection)
        False
        >>> issubscriptedcollectiontype(Mapping[str, str])
        True
        >>> issubscriptedcollectiontype(str)
        False
        >>> issubscriptedcollectiontype(NewType("Foo", Collection[int]))
        True
    """
    return iscollectiontype(obj) and issubscriptedgeneric(obj)


_ArgsT = tp.TypeVar("_ArgsT")


@compat.cache
def ismappingtype(obj: type) -> compat.TypeIs[type[tp.Mapping]]:
    """Test whether this annotation is a subtype of [`typing.Mapping`][].

    Examples:
        >>> from typing import Mapping, Dict, DefaultDict, NewType
        >>> ismappingtype(Mapping)
        True
        >>> ismappingtype(Dict[str, str])
        True
        >>> ismappingtype(DefaultDict)
        True
        >>> ismappingtype(dict)
        True
        >>> class MyDict(dict): ...
        ...
        >>> ismappingtype(MyDict)
        True
        >>> class MyMapping(Mapping): ...
        ...
        >>> ismappingtype(MyMapping)
        True
        >>> ismappingtype(NewType("Foo", dict))
        True
    """
    obj = origin(obj)
    return builtins.issubclass(obj, _MAPPING_TYPES) or builtins.issubclass(
        obj, tp.Mapping
    )


_MAPPING_TYPES = (dict, sqlite3.Row, types.MappingProxyType)


@compat.cache
def isenumtype(obj: type) -> compat.TypeIs[type[enum.Enum]]:
    """Test whether this annotation is a subclass of [`enum.Enum`][]

    Examples:
        >>> import enum
        >>>
        >>> class FooNum(enum.Enum): ...
        ...
        >>> isenumtype(FooNum)
        True
    """
    return issubclass(obj, enum.Enum)


@compat.cache
def isclassvartype(obj: type) -> bool:
    """Test whether an annotation is a ClassVar annotation.

    Examples:
        >>> from typing import ClassVar, NewType
        >>> isclassvartype(ClassVar[str])
        True
        >>> isclassvartype(NewType("Foo", ClassVar[str]))
        True
    """
    obj = resolve_supertype(obj)
    return getattr(obj, "__origin__", obj) is tp.ClassVar


_UNWRAPPABLE = (
    isclassvartype,
    isoptionaltype,
    isfinal,
)


@compat.cache
def should_unwrap(obj: type) -> bool:
    """Test whether we should use the __args__ attr for resolving the type.

    This is useful for determining what type to use at run-time for coercion.
    """
    return (not isliteral(obj)) and any(x(obj) for x in _UNWRAPPABLE)


@compat.cache
def isfromdictclass(obj: type) -> compat.TypeIs[type[_FromDict]]:
    """Test whether this annotation is a class with a `from_dict()` method."""
    return inspect.isclass(obj) and hasattr(obj, "from_dict")


class _FromDict(tp.Protocol):
    @classmethod
    def from_dict(cls: type[compat.Self], *args, **kwargs) -> compat.Self: ...


@compat.cache
def isfrozendataclass(obj: type) -> compat.TypeIs[type[_FrozenDataclass]]:
    """Test whether this is a dataclass and whether it's frozen."""
    return getattr(getattr(obj, "__dataclass_params__", None), "frozen", False)


class _FrozenDataclass(tp.Protocol):
    __dataclass_params__: dataclasses._DataclassParams  # type: ignore


cached_issubclass = compat.cache(issubclass)


__hashgetter = attrgetter("__hash__")


def ishashable(obj: tp.Any) -> compat.TypeIs[tp.Hashable]:
    """Check whether an object is hashable.

    An order of magnitude faster than [`isinstance`][] with
    [`typing.Hashable`][]

    Examples:
        >>> ishashable(str())
        True
        >>> ishashable(frozenset())
        True
        >>> ishashable(list())
        False
    """
    return __hashgetter(obj) is not None


@compat.cache
def istypeddict(obj: tp.Any) -> bool:
    """Check whether an object is a [`typing.TypedDict`][].

    Examples:
        >>> from typing import TypedDict
        >>>
        >>> class FooMap(TypedDict):
        ...     bar: str
        ...
        >>> istypeddict(FooMap)
        True
    """
    return (
        inspect.isclass(obj)
        and dict in {*inspect.getmro(obj)}
        and hasattr(obj, "__total__")
    )


@compat.cache
def istypedtuple(obj: type) -> compat.TypeIs[type[tp.NamedTuple]]:
    """Check whether an object is a "typed" tuple ([`typing.NamedTuple`][]).

    Examples:
        >>> from typing import NamedTuple
        >>>
        >>> class FooTup(NamedTuple):
        ...     bar: str
        ...
        >>> istypedtuple(FooTup)
        True
    """
    return (
        inspect.isclass(obj)
        and issubclass(obj, tuple)
        and bool(getattr(obj, "__annotations__", False))
    )


@compat.cache
def isnamedtuple(obj: type) -> compat.TypeIs[type[tp.NamedTuple]]:
    """Check whether an object is a "named" tuple ([`collections.namedtuple`][]).

    Examples:
        >>> from collections import namedtuple
        >>>
        >>> FooTup = namedtuple("FooTup", ["bar"])
        >>> isnamedtuple(FooTup)
        True
    """
    return inspect.isclass(obj) and issubclass(obj, tuple) and hasattr(obj, "_fields")


@compat.cache
def isfixedtupletype(obj: type) -> compat.TypeIs[type[tuple]]:
    """Check whether an object is a "fixed" tuple, e.g., `tuple[int, int]`.

    Examples:
        >>> from typing import Tuple
        >>>
        >>>
        >>> isfixedtupletype(Tuple[str, int])
        True
        >>> isfixedtupletype(Tuple[str, ...])
        False
    """
    a = args(obj)
    origin = tp.get_origin(obj)
    if not a or a[-1] is ...:
        return False
    return _safe_issubclass(origin, tuple)


def isforwardref(obj: tp.Any) -> compat.TypeIs[refs.ForwardRef]:
    """Tests whether the given object is a [`typing.ForwardRef`][]."""
    return obj.__class__ is refs.ForwardRef


def isproperty(obj) -> compat.TypeIs[types.DynamicClassAttribute]:
    """Test whether the given object is an instance of [`property`] or [`functools.cached_property`][].

    Examples:
        >>> import functools

        >>> class Foo:
        ...     @property
        ...     def prop(self) -> int:
        ...         return 1
        ...
        ...     @functools.cached_property
        ...     def cached(self) -> str:
        ...         return "foo"
        ...
        >>> isproperty(Foo.prop)
        True
        >>> isproperty(Foo.cached)
        True
    """

    return builtins.issubclass(obj.__class__, (property, functools.cached_property))


def isdescriptor(obj) -> compat.TypeIs[DescriptorT]:
    """Test whether the given object is a [`types.GetSetDescriptorType`][]

    Examples:
        >>> class StringDescriptor:
        ...     __slots__ = ("value",)
        ...
        ...     def __init__(self, default: str = "value"):
        ...         self.value = default
        ...
        ...     def __get__(self, instance: Any, value: str) -> str:
        ...         return self.value
        ...
        >>> isdescriptor(StringDescriptor)
        True
    """
    intersection = {*dir(obj)} & _DESCRIPTOR_METHODS
    return bool(intersection)


_DESCRIPTOR_METHODS = frozenset(("__get__", "__set__", "__delete__", "__set_name__"))


_VT_in = tp.TypeVar("_VT_in")
_VT_co = tp.TypeVar("_VT_co", covariant=True)
_VT_cont = tp.TypeVar("_VT_cont", contravariant=True)


class GetDescriptor(tp.Protocol[_VT_co]):
    @tp.overload
    def __get__(self, instance: None, owner: tp.Any) -> GetDescriptor: ...

    @tp.overload
    def __get__(self, instance: object, owner: tp.Any) -> _VT_co: ...

    def __get__(self, instance: tp.Any, owner: tp.Any) -> GetDescriptor | _VT_co: ...


class SetDescriptor(tp.Protocol[_VT_cont]):
    def __set__(self, instance: tp.Any, value: _VT_cont): ...


class DeleteDescriptor(tp.Protocol[_VT_co]):
    def __delete__(self, instance: tp.Any): ...


class SetNameDescriptor(tp.Protocol):
    def __set_name__(self, owner: tp.Any, name: str): ...


DescriptorT = tp.Union[
    GetDescriptor, SetDescriptor, DeleteDescriptor, SetNameDescriptor
]


def issimpleattribute(v) -> bool:
    """Test whether the given object is a static value

    (e.g., not a function, class, or descriptor).

    Examples:
        >>> class MyOperator:
        ...     type = str
        ...
        ...     def operate(self, v) -> type:
        ...         return self.type(v)
        ...
        ...     @property
        ...     def default(self) -> type:
        ...         return self.type()
        ...
        >>> issimpleattribute(MyOperator.type)
        False
        >>> issimpleattribute(MyOperator.operate)
        False
        >>> issimpleattribute(MyOperator.default)
        False
    """
    return not any(c(v) for c in _ATTR_CHECKS)


_ATTR_CHECKS = (inspect.isclass, inspect.isroutine, isproperty, isdescriptor)


def isabstract(o) -> compat.TypeIs[abc.ABC]:
    """Test whether the given object is an abstract type.

    Examples:
        >>> import abc
        >>> import numbers

        >>>
        >>> isabstract(numbers.Number)
        True
        >>>
        >>> class MyABC(abc.ABC): ...
        ...
        >>> isabstract(MyABC)
        True

    """
    return inspect.isabstract(o) or o in _ABCS


# Custom list of ABCs which incorrectly evaluate to false
_ABCS = frozenset({numbers.Number})


@compat.cache
def istexttype(t: type[tp.Any]) -> compat.TypeIs[type[str | bytes | bytearray]]:
    """Test whether the given type is a subclass of text or bytes.

    Examples:
        >>> class MyStr(str): ...
        ...
        >>> istexttype(MyStr)
        True
    """
    return _safe_issubclass(t, (str, bytes, bytearray, memoryview))


@compat.cache
def isstringtype(t: type[tp.Any]) -> compat.TypeIs[type[str | bytes | bytearray]]:
    """Test whether the given type is a subclass of text or bytes.

    Examples:
        >>> class MyStr(str): ...
        ...
        >>> istexttype(MyStr)
        True
    """
    return _safe_issubclass(t, str)


@compat.cache
def isbytestype(t: type[tp.Any]) -> compat.TypeIs[type[str | bytes | bytearray]]:
    """Test whether the given type is a subclass of text or bytes.

    Examples:
        >>> class MyStr(str): ...
        ...
        >>> istexttype(MyStr)
        True
    """
    return _safe_issubclass(t, (bytes, bytearray, memoryview))


@compat.cache
def isnumbertype(t: type[tp.Any]) -> compat.TypeIs[type[numbers.Number]]:
    """Test whether `t` is a subclass of the [`numbers.Number`][] protocol.

    Examples:
        >>> import decimal

        >>> isnumbertype(int)
        True
        >>> isnumbertype(float)
        True
        >>> isnumbertype(decimal.Decimal)
        True
    """
    return _safe_issubclass(t, numbers.Number)


@compat.cache
def isintegertype(t: type[tp.Any]) -> compat.TypeIs[type[int]]:
    """Test whether `t` is a subclass of the [`numbers.Number`][] protocol.

    Examples:
        >>> import decimal

        >>> isnumbertype(int)
        True
        >>> isnumbertype(float)
        False
        >>> isnumbertype(decimal.Decimal)
        False
    """
    return _safe_issubclass(t, int)


@compat.cache
def isfloattype(t: type[tp.Any]) -> compat.TypeIs[type[float]]:
    """Test whether `t` is a subclass of the [`numbers.Number`][] protocol.

    Examples:
        >>> import decimal

        >>> isnumbertype(int)
        False
        >>> isnumbertype(float)
        True
        >>> isnumbertype(decimal.Decimal)
        False
    """
    return _safe_issubclass(t, float)


@compat.cache
def isstructuredtype(t: type[tp.Any]) -> bool:
    """Test whether the given type has a fixed set of fields.

    Examples:
        >>> import dataclasses
        >>> from typing import Tuple, NamedTuple, TypedDict, Union, Literal, Collection

        >>>
        >>> isstructuredtype(Tuple[str, int])
        True
        >>> isstructuredtype(class MyDict(TypedDict): ...)
        True
        >>> isstructuredtype(class MyTup(NamedTuple): ...)
        True
        >>> isstructuredtype(class MyClass: ...)
        True
        >>> isstructuredtype(Union[str, int])
        False
        >>> isstructuredtype(Literal[1, 2])
        False
        >>> isstructuredtype(tuple)
        False
        >>> isstructuredtype(Collection[str])
        False
    """
    return (
        isfixedtupletype(t)
        or isnamedtuple(t)
        or istypeddict(t)
        or (not isstdlibsubtype(origin(t)) and not isuniontype(t) and not isliteral(t))
    )


@compat.cache
def isgeneric(t: tp.Any) -> bool:
    """Test whether the given type is a typing generic.

    Examples:
        >>> from typing import Tuple, Generic, TypeVar

        >>>
        >>> isgeneric(Tuple)
        True
        >>> isgeneric(tuple)
        False
        >>> T = TypeVar("T")
        >>> class MyGeneric(Generic[T]): ...
        >>> isgeneric(MyGeneric[int])
        True
    """
    strobj = str(t)
    is_generic = (
        strobj.startswith("typing.")
        or strobj.startswith("typing_extensions.")
        or "[" in strobj
        or _safe_issubclass(t, tp.Generic)  # type: ignore[arg-type]
    )
    return is_generic


@compat.cache
def issubscriptedgeneric(t: tp.Any) -> bool:
    """Test whether the given type is a typing generic.

    Examples:
        >>> from typing import Tuple, Generic, TypeVar

        >>>
        >>> isgeneric(Tuple)
        True
        >>> isgeneric(tuple)
        False
        >>> T = TypeVar("T")
        >>> class MyGeneric(Generic[T]): ...
        >>> isgeneric(MyGeneric[int])
        True
    """
    strobj = str(t)
    og = tp.get_origin(t) or t
    is_generic = isgeneric(og) or isgeneric(t)
    is_subscripted = "[" in strobj
    return is_generic and is_subscripted


@compat.cache  # type: ignore[arg-type]
def iscallable(t: tp.Any) -> compat.TypeIs[tp.Callable]:
    """Test whether the given type is a callable.

    Examples:
        >>> import typing
        >>> import collections.abc
        >>> iscallable(lambda: None)
        True
        >>> iscallable(typing.Callable)
        True
        >>> iscallable(collections.abc.Callable)
        True
        >>> iscallable(1)
        False
    """
    return inspect.isroutine(t) or t is tp.Callable or _safe_issubclass(t, abc_Callable)  # type: ignore[arg-type]


@compat.cache
def isunresolvable(t: tp.Any) -> bool:
    """Test whether the given type is unresolvable.

    Examples:
        >>> import typing
        >>> isunresolvable(int)
        False
        >>> isunresolvable(typ.Any)
        True
        >>> isunresolvable(...)
        True
    """
    return t in _UNRESOLVABLE


_UNRESOLVABLE = (
    object,
    tp.Any,
    re.Match,
    constants.empty,
    tp.Callable,
    abc_Callable,
    inspect.Parameter.empty,
    type(Ellipsis),
    Ellipsis,
)


@compat.cache
def isnonetype(t: tp.Any) -> compat.TypeIs[None]:
    """Detect if the given type is a [`types.NoneType`][].

    Examples:
        >>> isnonetype(None)
        True
        >>> isnonetype(type(None))
        True
        >>> isnonetype(1)
        False
    """
    return t in (None, type(None))


@compat.cache
def ispatterntype(t: tp.Any) -> compat.TypeIs[re.Pattern]:
    """Detect if the given type is a [`re.Pattern`][].

    Examples:
        >>> import re
        >>> ispatterntype(re.compile(r"^[a-z]+$"))
        True
        >>> ispatterntype(r"^[a-z]+$")
        False
    """
    return _safe_issubclass(t, re.Pattern)


@compat.cache
def ispathtype(t: tp.Any) -> compat.TypeIs[pathlib.Path]:
    """Detect if the given type is a [`pathlib.Path`][].

    Examples:
        >>> import pathlib
        >>> ispathtype(pathlib.Path.cwd())
        True
        >>> ispathtype(".")
        False
    """
    return _safe_issubclass(t, pathlib.PurePath)


@compat.cache
def istypealiastype(t: tp.Any) -> compat.TypeIs[compat.TypeAliasType]:
    """Detect if the given object is a [`typing.TypeAliasType`][].

    Examples:
        >>> type IntList = list[int]
        >>> istypealiastype(IntList)
        True
        >>> IntList = compat.TypeAliasType("IntList", list[int])
        >>> istypealiastype(IntList)
        True

    """
    return isinstance(t, compat.TypeAliasType)


def _safe_issubclass(__cls: type, __class_or_tuple: type | tuple[type, ...]) -> bool:
    try:
        return issubclass(__cls, __class_or_tuple)
    except TypeError:
        return False


# Here we are with a manually-defined set of builtin-types.
# This probably won't break anytime soon, but we shall see...
BuiltIntypeT = tp.Union[
    int, bool, float, str, bytes, bytearray, list, set, frozenset, tuple, dict, None
]
BUILTIN_TYPES = frozenset(
    (type(None), *(t for t in BuiltIntypeT.__args__ if t is not None))  # type: ignore
)
BUILTIN_TYPES_TUPLE = tuple(BUILTIN_TYPES)
STDLibtypeT = tp.Union[
    BuiltIntypeT,
    datetime.datetime,
    datetime.date,
    datetime.timedelta,
    datetime.time,
    decimal.Decimal,
    ipaddress.IPv4Address,
    ipaddress.IPv6Address,
    pathlib.Path,
    uuid.UUID,
    collections.defaultdict,
    collections.deque,
    types.MappingProxyType,
]
STDLIB_TYPES = frozenset(
    (type(None), *(t for t in STDLibtypeT.__args__ if t is not None))  # type: ignore
)
STDLIB_TYPES_TUPLE = tuple(STDLIB_TYPES)



from __future__ import annotations

import typing

__all__ = (
    "SQLAMetaData",
    "sqla_registry",
)


if typing.TYPE_CHECKING:

    class SQLAMetaData: ...

    class sqla_registry: ...


else:
    try:
        from sqlalchemy import MetaData as SQLAMetaData
    except (ImportError, ModuleNotFoundError):

        class SQLAMetaData: ...

    try:
        from sqlalchemy.orm import registry as sqla_registry
    except (ImportError, ModuleNotFoundError):

        class sqla_registry: ...



"""Utilities for working with stack traces and frames.

Examples: Typical Usage
    >>> import inspect
    >>> from typelib.py import frames
    >>> var = 1
    >>> frames.extract("var")
    1
    >>> current_frame = inspect.currentframe()
    >>> frames.getcaller() == current_frame
    True
"""

from __future__ import annotations

import inspect
import types
from typing import Any, Final

__all__ = (
    "extract",
    "getcaller",
    "PKG_NAME",
)


def extract(name: str, *, frame: types.FrameType = None) -> Any | None:
    """Extract `name` from the stacktrace of `frame`.

    If `frame` is not provided, this function will use the current frame.

    Args:
        name: The name of the object to extract from the stacktrace.
        frame: The [`types.FrameType`][] instance to start from (optional).
    """
    frame = frame or inspect.currentframe()
    seen: set[types.FrameType] = set()
    add = seen.add
    while frame and frame not in seen:
        if name in frame.f_globals:
            return frame.f_globals[name]
        if name in frame.f_locals:
            return frame.f_locals[name]
        add(frame)
        frame = frame.f_back

    return None


def getcaller(frame: types.FrameType = None) -> types.FrameType:
    """Get the caller of the current scope, excluding this library.

    If `frame` is not provided, this function will use the current frame.

    Args:
        frame: The [`types.FrameType`][] instance to start from (optional).
    """

    frame = frame or inspect.currentframe()
    while frame.f_back:
        frame = frame.f_back
        module = inspect.getmodule(frame)
        if module and module.__name__.startswith(PKG_NAME):
            continue

        code = frame.f_code
        if getattr(code, "co_qualname", "").startswith(PKG_NAME):
            continue
        if PKG_NAME in code.co_filename:
            continue
        return frame

    return frame


PKG_NAME: Final[str] = __name__.split(".", maxsplit=1)[0]



"""Constants used throughout the library."""

import typing as t


class empty:
    """A singleton for signalling no input."""


DEFAULT_ENCODING: t.Final[str] = "utf-8"
PKG_NAME: t.Final[str] = __name__.split(".", maxsplit=1)[0]



"""Utilities for automatic unmarshalling of inputs according to a callable object's signature.

Examples: Typical Usage
    >>> from typelib import binding
    >>>
    >>> def foo(val: int) -> int:
    ...     return val * 2
    ...
    >>> bound = binding.bind(foo)
    >>> bound("2")
    4
    >>> bound.call("3")
    '33'
"""

from __future__ import annotations

import abc
import dataclasses
import functools
import inspect
import typing as tp

from typelib import unmarshals
from typelib.py import classes, compat, inspection

P = compat.ParamSpec("P")
R = tp.TypeVar("R")
BindingT = compat.TypeAliasType(
    "BindingT",
    "tp.MutableMapping[str | int, unmarshals.AbstractUnmarshaller]",
)


def bind(obj: tp.Callable[P, R]) -> BoundRoutine[P, R]:
    """Create a type-enforced, bound routine for a callable object.

    Note:
        In contrast to [`typelib.binding.wrap`][], this function creates a new,
        type-enforced [`BoundRoutine`][typelib.binding.BoundRoutine] instance. Rather than
        masquerading as the given `obj`, we encapsulate it in the routine
        instance, which is more obvious and provides developers with the ability to
        side-step type enforcement when it is deemed unnecessary, which should be
        most of the time if your code is strongly typed and statically analyzed.

    Tip: TL;DR
        This function returns an object that walks like your duck and quacks like your duck,
        but doesn't look like your duck.

    Args:
        obj: The callable object to bind.
    """
    binding: AbstractBinding[P] = _get_binding(obj)
    routine: BoundRoutine[P, R] = BoundRoutine(
        call=obj,
        binding=binding,
    )
    return routine


def wrap(obj: tp.Callable[P, R]) -> tp.Callable[..., R]:
    """Wrap a callable object for runtime type coercion of inputs.

    Note:
        If a class is given, we will attempt to wrap the init method.

    Warning:
        This is a useful feature. It is also very *surprising*. By wrapping a callable
        in this decorator, we end up with *implicit* behavior that's not obvious to the
        caller or a fellow developer.

        You're encouraged to prefer [`typelib.binding.bind`][] for similar
        functionality, less the implicit nature, especially when a class is given.

    Tip: TL;DR
        This function returns an object walks, quacks, and tries to look like your duck.

    Args:
        obj: The callable object to wrap.
             Maybe be a function, a callable class instance, or a class.
    """

    binding: AbstractBinding[P] = _get_binding(obj)

    if inspect.isclass(obj):
        obj.__init__ = wrap(obj.__init__)
        return obj

    @functools.wraps(obj)  # type: ignore[arg-type]
    def binding_wrapper(*args: tp.Any, __binding=binding, **kwargs: tp.Any) -> R:
        bargs, bkwargs = __binding(args, kwargs)
        return obj(*bargs, **bkwargs)

    return binding_wrapper


@classes.slotted(dict=False, weakref=True)
@dataclasses.dataclass
class BoundRoutine(tp.Generic[P, R]):
    """A type-enforced, bound routine for a callable object."""

    call: tp.Callable[P, R]
    """The callable object."""
    binding: AbstractBinding[P]
    """The parameter->type binding."""

    def __call__(self, *args: tp.Any, **kwargs: tp.Any) -> R:
        """Binding an input to the parameters of `call`,

        then call the callable and return the result."""
        bargs, bkwargs = self.binding(args=args, kwargs=kwargs)
        return self.call(*bargs, **bkwargs)


@compat.cache
def _get_binding(obj: tp.Callable) -> AbstractBinding:
    sig = inspection.cached_signature(obj)
    params = sig.parameters
    binding: BindingT = {}
    has_pos_only = False
    has_kwd_only = False
    has_args = False
    has_kwargs = False
    has_pos_or_kwd = False
    max_pos: int | None = None
    varkwd: unmarshals.AbstractUnmarshaller | None = None
    varpos: unmarshals.AbstractUnmarshaller | None = None
    for i, (name, param) in enumerate(params.items()):
        unmarshaller: unmarshals.AbstractUnmarshaller = unmarshals.unmarshaller(
            param.annotation
        )
        binding[name] = binding[i] = unmarshaller
        has_kwd_only = has_kwd_only or param.kind == inspect.Parameter.KEYWORD_ONLY
        has_pos_or_kwd = (
            has_pos_or_kwd or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        )
        if param.kind == param.POSITIONAL_ONLY:
            has_pos_only = True
            max_pos = i
            continue
        if param.kind == param.VAR_KEYWORD:
            has_kwargs = True
            varkwd = unmarshaller
            continue
        if param.kind == param.VAR_POSITIONAL:
            has_args = True
            max_pos = i - 1
            varpos = unmarshaller

    truth = _Truth(
        has_pos_only=has_pos_only,
        has_kwd_only=has_kwd_only,
        has_args=has_args,
        has_kwargs=has_kwargs,
        has_pos_or_kwd=has_pos_or_kwd,
    )
    binding_cls = _BINDING_CLS_MATRIX[truth]
    return binding_cls(
        binding=binding,
        signature=sig,
        varkwd=varkwd,
        varpos=varpos,
        startpos=max_pos + 1 if max_pos is not None else None,
    )


class AbstractBinding(abc.ABC, tp.Generic[P]):
    """The abstract base class for all type-enforced bindings.

    Note:
        "Bindings" are callables which leverage the type annotations in a signature to
        unmarshal inputs.

        We differentiate each subclass based upon the possible combinations of parameter
        kinds:

        - Positional-only arguments
        - Keyword-only arguments
        - Positional-or-Keyword arguments
        - Variable-positional arguments (`*args`)
        - Variable-keyword arguments (`**kwargs`)

        This allows us to micro-optimize the call for each subclass to exactly what is
        necessary for the that combination, which can lead to a significant speedup in
        hot loops.
    """

    __slots__ = ("binding", "signature", "varkwd", "varpos", "startpos")

    def __init__(
        self,
        *,
        signature: inspect.Signature,
        binding: BindingT,
        varkwd: unmarshals.AbstractUnmarshaller | None = None,
        varpos: unmarshals.AbstractUnmarshaller | None = None,
        startpos: int | None = None,
    ):
        """Constructor.

        Args:
            signature: The signature for the binding.
            binding: A mapping of parameter names and positions to unmarshallers.
                     This accounts for positional, keyword, or positional-or-keyword arguments.
            varkwd: The unmarshaller for var-keyword arguments (`**kwargs`).
            varpos: The unmarshaller for var-positional arguments (`*args`).
            startpos: The start position of var-positional arguments (`*args`).
                      This accounts for the fact that var-positional comes after positional-only.
        """
        self.signature = signature
        self.binding = binding
        self.varkwd = varkwd
        self.varpos = varpos
        self.startpos = startpos

    def __repr__(self):
        return f"<{self.__class__.__name__}(signature={self.signature})>"

    @abc.abstractmethod
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        """Inspect the given `args` and `kwargs` and unmarshal them.

        Args:
            args: The positional arguments.
            kwargs: The keyword arguments.
        """


class AnyParamKindBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Split the supplied args into the positional and the var-args
        #   Implementation note: if there are positional args and var-args,
        #   positional args will always be first.
        posargs = args[: self.startpos]
        varargs = args[self.startpos :]
        # Unmarshal the positional arguments.
        umargs = (
            *(binding[i](v) if i in binding else v for i, v in enumerate(posargs)),
            *(varpos(v) for v in varargs),
        )
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding.get(k, varkwd)(v) for k, v in kwargs.items()}
        return umargs, umkwargs


class PosArgsKwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Split the supplied args into the positional and the var-args
        #   Implementation note: if there are positional args and var-args,
        #   positional args will always be first.
        posargs = args[: self.startpos]
        varargs = args[self.startpos :]
        umargs = (
            *(binding[i](v) if i in binding else v for i, v in enumerate(posargs)),
            *(varpos(v) for v in varargs),
        )
        umkwargs = {k: varkwd(v) for k, v in kwargs.items()}
        return umargs, umkwargs


class PosKwdKwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Unmarshal the args
        umargs = (*(binding[i](v) if i in binding else v for i, v in enumerate(args)),)
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding.get(k, varkwd)(v) for k, v in kwargs.items()}
        return umargs, umkwargs


class PosKwdArgsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        # Split the supplied args into the positional and the var-args
        #   Implementation note: if there are positional args and var-args,
        #   positional args will always be first.
        posargs = args[: self.startpos]
        varargs = args[self.startpos :]
        # Unmarshal the positional arguments.
        umargs = (
            *(binding[i](v) if i in binding else v for i, v in enumerate(posargs)),
            *(varpos(v) for v in varargs),
        )
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding[k](v) if k in binding else v for k, v in kwargs.items()}
        return umargs, umkwargs


class PosKwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Unmarshal the args
        umargs = (*(binding[i](v) if i in binding else v for i, v in enumerate(args)),)
        # Unmarshal the keyword arguments.
        umkwargs = {k: varkwd(v) for k, v in kwargs.items()}
        return umargs, umkwargs


class PosKwdBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        # Unmarshal the args
        umargs = (*(binding[i](v) if i in binding else v for i, v in enumerate(args)),)
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding[k](v) if k in binding else k for k, v in kwargs.items()}
        return umargs, umkwargs


class PosArgsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        # Split the supplied args into the positional and the var-args
        #   Implementation note: if there are positional args and var-args,
        #   positional args will always be first.
        posargs = args[: self.startpos]
        varargs = args[self.startpos :]
        umargs = (
            *(binding[i](v) if i in binding else v for i, v in enumerate(posargs)),
            *(varpos(v) for v in varargs),
        )
        return umargs, kwargs


class PosBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize key attributes
        binding = self.binding
        # Unmarshal the args
        umargs = (*(binding[i](v) if i in binding else v for i, v in enumerate(args)),)
        return umargs, kwargs


class KwdArgsKwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize the key attributes
        binding = self.binding
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Unmarshal the positional args.
        umargs = (*(varpos(v) for v in args),)
        # Unmarshal the keyword args.
        umkwargs = {k: binding.get(k, varkwd)(v) for k, v in kwargs.items()}
        return umargs, umkwargs


class KwdArgsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize the key attributes
        binding = self.binding
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        # Unmarshal the positional arguments
        umargs = (*(varpos(v) for v in args),)
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding[k](v) if k in binding else k for k, v in kwargs.items()}
        return umargs, umkwargs


class KwdKwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        binding = self.binding
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding.get(k, varkwd)(v) for k, v in kwargs.items()}
        return args, umkwargs


class KwdBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        binding = self.binding
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding[k](v) if k in binding else k for k, v in kwargs.items()}
        return args, umkwargs


class ArgsKwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize the key attributes
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Unmarshal the positional arguments.
        umargs = (*(varpos(v) for v in args),)
        # Unmarshal the keyword arguments.
        umkwargs = {k: varkwd(v) for k, v in kwargs.items()}
        return umargs, umkwargs


class KwargsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize the key attributes
        varkwd: unmarshals.AbstractUnmarshaller = self.varkwd
        # Unmarshal the keyword arguments.
        umkwargs = {k: varkwd(v) for k, v in kwargs.items()}
        return args, umkwargs


class ArgsBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize the key attributes
        varpos: unmarshals.AbstractUnmarshaller = self.varpos
        # Unmarshal the positional arguments.
        umargs = (*(varpos(v) for v in args),)
        return umargs, kwargs


class PosOrKwdBinding(AbstractBinding[P], tp.Generic[P]):
    def __call__(
        self, args: tuple[tp.Any], kwargs: dict[str, tp.Any]
    ) -> tuple[P.args, P.kwargs]:
        # Localize the key attributes
        binding = self.binding
        # Unmarshal the positional args.
        umargs = (*(binding[i](v) if i in binding else v for i, v in enumerate(args)),)
        # Unmarshal the keyword arguments.
        umkwargs = {k: binding[k](v) if k in binding else k for k, v in kwargs.items()}
        return umargs, umkwargs


class _Truth(tp.NamedTuple):
    """Internal truth entry for determining the binding algorithm.

    Attributes:
        has_pos_only: Whether the signature contains positional-only arguments.
        has_kwd_only: Whether the signature contains keyword-only arguments.
        has_args: Whether the signature contains variadic-positional arguments.
        has_kwargs: Whether the signature contains variadic-keyword arguments.
        has_pos_or_kwd: Whether the signature contains positional-or-keyword arguments.
    """

    has_pos_only: bool = False
    has_kwd_only: bool = False
    has_args: bool = False
    has_kwargs: bool = False
    has_pos_or_kwd: bool = False


_BINDING_CLS_MATRIX: dict[_Truth, type[AbstractBinding]] = {
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): PosOrKwdBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosOrKwdBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): KwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): PosKwdKwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): ArgsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosArgsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): ArgsKwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): ArgsKwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): KwdBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosOrKwdBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): KwdKwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): PosKwdKwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): KwdArgsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosOrKwdBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): KwdArgsKwargsBinding,
    _Truth(
        has_pos_only=False,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): KwdArgsKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): PosBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosOrKwdBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): PosKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): PosKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): PosArgsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosArgsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): PosArgsKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=False,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): AnyParamKindBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): PosKwdBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosKwdKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): PosArgsKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=False,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): PosKwdKwargsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=False,
    ): PosKwdArgsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=False,
        has_pos_or_kwd=True,
    ): PosKwdArgsBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=False,
    ): AnyParamKindBinding,
    _Truth(
        has_pos_only=True,
        has_kwd_only=True,
        has_args=True,
        has_kwargs=True,
        has_pos_or_kwd=True,
    ): AnyParamKindBinding,
}



"""The API for unmarshalling simple Python objects into higher-order Python types."""

from __future__ import annotations

import typing as tp

from typelib import ctx, graph
from typelib.py import compat, inspection, refs
from typelib.unmarshals import routines

T = tp.TypeVar("T")

__all__ = (
    "unmarshal",
    "unmarshaller",
    "DelayedUnmarshaller",
)


def unmarshal(t: type[T] | refs.ForwardRef | str, value: tp.Any) -> T:
    """Unmarshal `value` into `typ`.

    Args:
        t: The type annotation or reference to unmarshal into.
        value: The value to unmarshal.
    """
    routine = unmarshaller(t)
    unmarshalled = routine(value)
    return unmarshalled


@compat.cache
def unmarshaller(
    t: type[T] | refs.ForwardRef | compat.TypeAliasType | str,
) -> routines.AbstractUnmarshaller[T]:
    """Get an un-marshaller routine for a given type.

    Args:
        t: The type annotation to generate an unmarshaller for.
             May be a type, type alias, [`typing.ForwardRef`][], or string reference.
    """
    nodes = graph.static_order(t)
    context: dict[type | graph.TypeNode, routines.AbstractUnmarshaller] = (
        ctx.TypeContext()
    )
    if not nodes:
        return routines.NoOpUnmarshaller(t=t, context=context, var=None)  # type: ignore[arg-type]

    # "root" type will always be the final node in the sequence.
    root = nodes[-1]
    for node in nodes:
        context[node] = _get_unmarshaller(node, context=context)

    return context[root]


def _get_unmarshaller(  # type: ignore[return]
    node: graph.TypeNode,
    context: routines.ContextT,
) -> routines.AbstractUnmarshaller[T]:
    if node in context:
        return context[node]

    for check, unmarshaller_cls in _HANDLERS.items():
        if check(node.type):
            return unmarshaller_cls(node.type, context=context, var=node.var)

    return routines.StructuredTypeUnmarshaller(node.type, context=context, var=node.var)


class DelayedUnmarshaller(routines.AbstractUnmarshaller[T]):
    """Delayed proxy for a given type's unmarshaller, used when we encounter a [`typing.ForwardRef`][].

    Notes:
        This allows us to delay the resolution of the given type reference until
        call-time, enabling support for cyclic and recursive types.
    """

    def __init__(
        self, t: type[T], context: routines.ContextT, *, var: str | None = None
    ):
        super().__init__(t, context, var=var)
        self._resolved: routines.AbstractUnmarshaller[T] | None = None

    @property
    def resolved(self) -> routines.AbstractUnmarshaller[T]:
        """The resolved unmarshaller."""
        if self._resolved is None:
            self._resolved = unmarshaller(self.t)
            for attr in self._resolved.__slots__:
                setattr(self, attr, getattr(self._resolved, attr))
        return self._resolved

    def __call__(self, val: tp.Any) -> T:
        unmarshalled = self.resolved(val)
        return unmarshalled


# Order is IMPORTANT! This is a FIFO queue.
_HANDLERS: tp.Mapping[
    tp.Callable[[type[T]], bool], type[routines.AbstractUnmarshaller]
] = {
    # Short-circuit forward refs
    inspection.isforwardref: DelayedUnmarshaller,
    inspection.isunresolvable: routines.NoOpUnmarshaller,
    inspection.isnonetype: routines.NoneTypeUnmarshaller,
    # Special handler for Literals
    inspection.isliteral: routines.LiteralUnmarshaller,
    # Special handler for Unions...
    inspection.isuniontype: routines.UnionUnmarshaller,
    # Non-intersecting types (order doesn't matter here.
    inspection.isdatetimetype: routines.DateTimeUnmarshaller,
    inspection.isdatetype: routines.DateUnmarshaller,
    inspection.istimetype: routines.TimeUnmarshaller,
    inspection.istimedeltatype: routines.TimeDeltaUnmarshaller,
    inspection.isuuidtype: routines.UUIDUnmarshaller,
    inspection.ispatterntype: routines.PatternUnmarshaller,
    inspection.ispathtype: routines.PathUnmarshaller,
    inspection.isdecimaltype: routines.DecimalUnmarshaller,
    inspection.isfractiontype: routines.FractionUnmarshaller,
    inspection.isnumbertype: routines.NumberUnmarshaller,
    inspection.isstringtype: routines.StringUnmarshaller,
    inspection.isbytestype: routines.BytesUnmarshaller,
    # Psuedo-structured containers, should check before generics.
    inspection.istypeddict: routines.StructuredTypeUnmarshaller,
    inspection.istypedtuple: routines.StructuredTypeUnmarshaller,
    inspection.isnamedtuple: routines.StructuredTypeUnmarshaller,
    inspection.isfixedtupletype: routines.FixedTupleUnmarshaller,
    (
        lambda t: inspection.issubscriptedgeneric(t) and inspection.ismappingtype(t)
    ): routines.SubscriptedMappingUnmarshaller,
    (
        lambda t: inspection.issubscriptedgeneric(t) and inspection.isiteratortype(t)
    ): routines.SubscriptedIteratorUnmarshaller,
    (
        lambda t: inspection.issubscriptedgeneric(t) and inspection.isiterabletype(t)
    ): routines.SubscriptedIterableUnmarshaller,
    # A mapping is a collection so must come before that check.
    inspection.ismappingtype: routines.MappingUnmarshaller,
    # Generic iterator handler
    inspection.isiteratortype: routines.NoOpUnmarshaller,
    # Generic Iterable handler
    inspection.isiterabletype: routines.IterableUnmarshaller,
}



"""Type-specific logic for unmarshalling simple Python objects into higher-order Python types."""

from __future__ import annotations

import abc
import contextlib
import datetime
import decimal
import fractions
import numbers
import pathlib
import re
import types
import typing as tp
import uuid

from typelib import constants, graph, serdes
from typelib.py import compat, inspection

T = tp.TypeVar("T")

__all__ = (
    "AbstractUnmarshaller",
    "ContextT",
    "NoOpUnmarshaller",
    "NoneTypeUnmarshaller",
    "BytesUnmarshaller",
    "StringUnmarshaller",
    "NumberUnmarshaller",
    "DecimalUnmarshaller",
    "FractionUnmarshaller",
    "DateUnmarshaller",
    "DateTimeUnmarshaller",
    "TimeUnmarshaller",
    "TimeDeltaUnmarshaller",
    "UUIDUnmarshaller",
    "PathUnmarshaller",
    "CastUnmarshaller",
    "PatternUnmarshaller",
    "MappingUnmarshaller",
    "IterableUnmarshaller",
    "LiteralUnmarshaller",
    "UnionUnmarshaller",
    "SubscriptedIteratorUnmarshaller",
    "SubscriptedIterableUnmarshaller",
    "SubscriptedMappingUnmarshaller",
    "FixedTupleUnmarshaller",
    "StructuredTypeUnmarshaller",
)


class AbstractUnmarshaller(abc.ABC, tp.Generic[T]):
    """Abstract base class defining the common interface for unmarshallers.

    Unmarshallers are custom callables which maintain type-specific information. They
    use this information to provide robust, performant logic for decoding and converting
    primtive Python objects or JSON-endcoded data into their target type.

    Unmarshallers support contextual deserialization, which enables the unmarshalling of
    nested types.

    Attributes:
        t: The root type of this unmarshaller.
        origin: If `t` is a generic, this will be an actionable runtime type
                related to `t`, otherwise it is the same as `t`.
        context: The complete type context for this unmarshaller.
        var: If this unmarshaller is used in a nested context, this will reference the
             field/parameter/index at which this unmarshaller should be used.
    """

    t: type[T]
    origin: type[T]
    context: ContextT
    var: str | None

    __slots__ = ("t", "origin", "context", "var")

    def __repr__(self):
        return f"<{self.__class__.__name__}(type={self.t!r}, origin={self.origin!r}, var={self.var!r})>"

    def __init__(self, t: type[T], context: ContextT, *, var: str | None = None):
        """Construct an unmarshaller instance.

        Args:
            t: The root type of this unmarshaller.
            context: The complete type context for this unmarshaller.
            var: The associated field or parameter name for this unmarshaller (optional).
        """
        self.t = t
        self.origin = inspection.origin(self.t)
        self.context = context
        self.var = var

    @abc.abstractmethod
    def __call__(self, val: tp.Any) -> T:
        """Unmarshall a Python object into its target type.

        Not implemented for the abstract base class.
        """


class NoOpUnmarshaller(AbstractUnmarshaller[T]):
    """Unmarshaller that does nothing."""

    def __call__(self, val: tp.Any) -> T:
        return tp.cast(T, val)


class NoneTypeUnmarshaller(AbstractUnmarshaller[None]):
    """Unmarshaller for null values.

    Note:
        We will attempt to decode any string/bytes input before evaluating for `None`.

    See Also:
        - [`typelib.serdes.decode`][]
    """

    def __call__(self, val: tp.Any) -> None:
        """Unmarshal the given input into a `None` value.

        Args:
            val: The value to unmarshal.

        Raises:
            ValueError: If `val` is not `None` after decoding.
        """
        decoded = serdes.decode(val)
        if decoded is not None:
            raise ValueError(f"{val!r} is not of {types.NoneType!r}")
        return None


ContextT: tp.TypeAlias = "tp.Mapping[type | graph.TypeNode, AbstractUnmarshaller]"
BytesT = tp.TypeVar("BytesT", bound=bytes)


class BytesUnmarshaller(AbstractUnmarshaller[BytesT], tp.Generic[BytesT]):
    """Unmarshaller that encodes an input to bytes.

    Note:
        We will format a member of the `datetime` module into ISO format before converting to bytes.

    See Also:
        - [`typelib.serdes.isoformat`][]
    """

    def __call__(self, val: tp.Any) -> BytesT:
        if isinstance(val, self.t):
            return val
        # Always encode date/time as ISO strings.
        if isinstance(val, (datetime.date, datetime.time, datetime.timedelta)):
            val = serdes.isoformat(val)
        return self.t(str(val).encode(constants.DEFAULT_ENCODING))


StringT = tp.TypeVar("StringT", bound=str)


class StringUnmarshaller(AbstractUnmarshaller[StringT], tp.Generic[StringT]):
    """Unmarshaller that converts an input to a string.

    Note:
        We will format a member of the `datetime` module into ISO format.

    See Also:
        - [`typelib.serdes.isoformat`][]
    """

    def __call__(self, val: tp.Any) -> StringT:
        # Always decode bytes.
        decoded = serdes.decode(val)
        if isinstance(decoded, self.t):
            return decoded
        # Always encode date/time as ISO strings.
        if isinstance(val, (datetime.date, datetime.time, datetime.timedelta)):
            decoded = serdes.isoformat(val)
        return self.t(decoded)


NumberT = tp.TypeVar("NumberT", bound=numbers.Number)


class NumberUnmarshaller(AbstractUnmarshaller[NumberT], tp.Generic[NumberT]):
    """Unmarshaller that converts an input to a number.

    Note:
        Number unmarshalling follows a best-effort strategy. We may extend type resolution
        to support more advanced type unmarshalling strategies in the future.

        As of now:
            1. Attempt to decode any bytes/string input into a real Python value.
            2. If the input is a member of the `datetime` module, convert it to a number.
            3. If the input is a mapping, unpack it into the number constructor.
            4. If the input is an iterable, unpack it into the number constructor.
            5. Otherwise, call the number constructor with the input.

    See Also:
        - [`typelib.serdes.unixtime`][]
    """

    def __call__(self, val: tp.Any) -> NumberT:
        """Unmarshall a value into the bound Number type.

        Args:
            val: The input value to unmarshal.
        """
        # Always decode bytes.
        decoded = serdes.decode(val)
        if isinstance(decoded, self.t):
            return decoded
        # Represent date/time objects as time since unix epoch.
        if isinstance(val, (datetime.date, datetime.time, datetime.timedelta)):
            decoded = serdes.unixtime(val)
        # Treat containers as constructor args.
        if inspection.ismappingtype(decoded.__class__):
            return self.t(**decoded)
        if inspection.isiterabletype(decoded.__class__) and not inspection.istexttype(
            decoded.__class__
        ):
            return self.t(*decoded)
        # Simple cast for non-containers.
        return self.t(decoded)  # type: ignore[call-arg]


DecimalT = tp.TypeVar("DecimalT", bound=decimal.Decimal)
DecimalUnmarshaller = NumberUnmarshaller[DecimalT]

FractionT = tp.TypeVar("FractionT", bound=fractions.Fraction)
FractionUnmarshaller = NumberUnmarshaller[FractionT]


DateT = tp.TypeVar("DateT", bound=datetime.date)


class DateUnmarshaller(AbstractUnmarshaller[DateT], tp.Generic[DateT]):
    """Unmarshaller that converts an input to a [`datetime.date`][] (or subclasses).

    Notes:
        This class tries to handle the 90% case:

        1. If we are already a [`datetime.date`][] instance, return it.
        2. If we are a `float` or `int` instance, treat it as a unix timestamp, at UTC.
        3. Attempt to decode any bytes/string input into a real Python value.
        4. If we have a string value, parse it into either a [`datetime.date`][]
        5. If the parsed result is a [`datetime.time`][] instance, then return
           the result of [`datetime.datetime.now`][], at UTC, as a [`datetime.date`][].

    Tip: TL;DR
        There are many ways to represent a date object over-the-wire. Your most
        fool-proof method is to rely upon [ISO 8601][iso] or [RFC 3339][rfc].

        [iso]: https://en.wikipedia.org/wiki/ISO_8601
        [rfc]: https://tools.ietf.org/html/rfc3339

    See Also:
        - [`typelib.serdes.decode`][]
        - [`typelib.serdes.dateparse`][]

    """

    def __call__(self, val: tp.Any) -> DateT:
        """Unmarshal a value into the bound `DateT` type.

        Args:
            val: The input value to unmarshal.
        """
        if isinstance(val, self.t) and not isinstance(val, datetime.datetime):
            return val

        # Numbers can be treated as time since epoch.
        if isinstance(val, (int, float)):
            val = datetime.datetime.fromtimestamp(val, tz=datetime.timezone.utc)
        # Always decode bytes.
        decoded = serdes.decode(val)
        # Parse strings.
        date: datetime.date | datetime.time = (
            serdes.dateparse(decoded, self.t) if isinstance(decoded, str) else decoded
        )
        # Time-only construct is treated as today.
        if isinstance(date, datetime.time):
            date = datetime.datetime.now(tz=datetime.timezone.utc).today()
        # Exact class matching - the parser returns subclasses.
        if date.__class__ is self.t:
            return date  # type: ignore[return-value]
        # Reconstruct as the exact type.
        return self.t(year=date.year, month=date.month, day=date.day)


DateTimeT = tp.TypeVar("DateTimeT", bound=datetime.datetime)


class DateTimeUnmarshaller(
    AbstractUnmarshaller[datetime.datetime], tp.Generic[DateTimeT]
):
    """Unmarshaller that converts an input to a [`datetime.datetime`][] (or subclasses).

    Notes:
        This class tries to handle the 90% case:

        1. If we are already a [`datetime.datetime`][] instance, return it.
        2. If we are a `float` or `int` instance, treat it as a unix timestamp, at UTC.
        3. Attempt to decode any bytes/string input into a real Python value.
        4. If we have a string value, parse it into either a [`datetime.date`][]
           instance, a [`datetime.time`][] instance or a [`datetime.datetime`][].
        5. If the parsed result is a [`datetime.time`][] instance, then merge
           the parsed time with today, at the timezone specified in the time instance.
        6. If the parsed result is a [`datetime.date`][] instance, create a
           [`datetime.datetime`][] instance at midnight of the indicated date, UTC.

    Tip: TL;DR
        There are many ways to represent a datetime object over-the-wire. Your most
        fool-proof method is to rely upon [ISO 8601][iso] or [RFC 3339][rfc].

        [iso]: https://en.wikipedia.org/wiki/ISO_8601
        [rfc]: https://tools.ietf.org/html/rfc3339


    See Also:
        - [`typelib.serdes.decode`][]
        - [`typelib.serdes.dateparse`][]


    """

    def __call__(self, val: tp.Any) -> datetime.datetime:
        """Unmarshal a value into the bound `DateTimeT` type.

        Args:
            val: The input value to unmarshal.
        """
        if isinstance(val, self.t):
            return val

        # Numbers can be treated as time since epoch.
        if isinstance(val, (int, float)):
            val = datetime.datetime.fromtimestamp(val, tz=datetime.timezone.utc)
        # Always decode bytes.
        decoded = serdes.decode(val)
        # Parse strings.
        dt: datetime.datetime | datetime.date | datetime.time = (
            serdes.dateparse(decoded, self.t) if isinstance(decoded, str) else decoded
        )
        # If we have a time object, default to today.
        if isinstance(dt, datetime.time):
            return self.t.now(tz=dt.tzinfo).replace(
                hour=dt.hour,
                minute=dt.minute,
                second=dt.second,
                microsecond=dt.microsecond,
                tzinfo=dt.tzinfo,
            )
        # Exact class matching.
        if dt.__class__ is self.t:
            return dt  # type: ignore[return-value]
        # Subclass check for datetimes.
        if isinstance(dt, datetime.datetime):
            return self.t(
                year=dt.year,
                month=dt.month,
                day=dt.day,
                hour=dt.hour,
                minute=dt.minute,
                second=dt.second,
                microsecond=dt.microsecond,
                tzinfo=dt.tzinfo,
                fold=dt.fold,
            )
        # Implicit: we have a date object.
        return self.t(
            year=dt.year, month=dt.month, day=dt.day, tzinfo=datetime.timezone.utc
        )


TimeT = tp.TypeVar("TimeT", bound=datetime.time)


class TimeUnmarshaller(AbstractUnmarshaller[TimeT], tp.Generic[TimeT]):
    """Unmarshaller that converts an input to a[`datetime.time`][] (or subclasses).

    Notes:
        This class tries to handle the 90% case:

        1. If we are already a [`datetime.time`][] instance, return it.
        2. If we are a `float` or `int` instance, treat it as a unix timestamp, at UTC.
        3. Attempt to decode any bytes/string input into a real Python value.
        4. If we have a string value, parse it into either a [`datetime.date`][]
           instance, a [`datetime.time`][] instance or a [`datetime.datetime`][].
        5. If the parsed result is a [`datetime.datetime`][] instance, then
           extract the time portion, preserving the associated timezone.
        6. If the parsed result is a [`datetime.date`][] instance, create
           a time instance at midnight, UTC.

    Tip: TL;DR
        There are many ways to represent a time object over-the-wire. Your most
        fool-proof method is to rely upon [ISO 8601][iso] or [RFC 3339][rfc].

        [iso]: https://en.wikipedia.org/wiki/ISO_8601
        [rfc]: https://tools.ietf.org/html/rfc3339

    See Also:
        - [`typelib.serdes.decode`][]
        - [`typelib.serdes.dateparse`][]

    """

    def __call__(self, val: tp.Any) -> TimeT:
        """Unmarshal a value into the bound `TimeT` type.

        Args:
            val: The input value to unmarshal.
        """
        if isinstance(val, self.t):
            return val

        decoded = serdes.decode(val)
        if isinstance(decoded, (int, float)):
            decoded = (
                datetime.datetime.fromtimestamp(val, tz=datetime.timezone.utc)
                .time()
                # datetime.time() strips tzinfo...
                .replace(tzinfo=datetime.timezone.utc)
            )
        dt: datetime.datetime | datetime.date | datetime.time = (
            serdes.dateparse(decoded, self.t) if isinstance(decoded, str) else decoded
        )

        if isinstance(dt, datetime.datetime):
            # datetime.time() strips tzinfo...
            dt = dt.time().replace(tzinfo=dt.tzinfo)
        elif isinstance(dt, datetime.date):
            dt = self.t(tzinfo=datetime.timezone.utc)

        if dt.__class__ is self.t:
            return dt  # type: ignore[return-value]

        return self.t(
            hour=dt.hour,
            minute=dt.minute,
            second=dt.second,
            microsecond=dt.microsecond,
            tzinfo=dt.tzinfo,
            fold=dt.fold,
        )


TimeDeltaT = tp.TypeVar("TimeDeltaT", bound=datetime.timedelta)


class TimeDeltaUnmarshaller(AbstractUnmarshaller[TimeDeltaT], tp.Generic[TimeDeltaT]):
    """Unmarshaller that converts an input to a [`datetime.timedelta`][] (or subclasses).

    Notes:
        This class tries to handle the 90% case:

        1. If we are already a [`datetime.timedelta`][] instance, return it.
        2. If we are a `float` or `int` instance, treat it as total seconds for a delta.
        3. Attempt to decode any bytes/string input into a real Python value.
        4. If we have a string value, parse it into a [`datetime.timedelta`][] instance.
        5. If the parsed result is not *exactly* the bound `TimeDeltaT` type, convert it.


    Tips: TL;DR
        There are many ways to represent a time object over-the-wire. Your most
        fool-proof method is to rely upon [ISO 8601][iso] or [RFC 3339][rfc].

        [iso]: https://en.wikipedia.org/wiki/ISO_8601
        [rfc]: https://tools.ietf.org/html/rfc3339


    See Also:
        - [`typelib.serdes.decode`][]
        - [`typelib.serdes.dateparse`][]

    """

    def __call__(self, val: tp.Any) -> TimeDeltaT:
        """Unmarshal a value into the bound `TimeDeltaT` type.

        Args:
            val: The input value to unmarshal.
        """
        if isinstance(val, (int, float)):
            return self.t(seconds=int(val))

        decoded = serdes.decode(val)
        td: datetime.timedelta = (
            serdes.dateparse(decoded, t=datetime.timedelta)
            if isinstance(decoded, str)
            else decoded
        )

        if td.__class__ is self.t:
            return td  # type: ignore[return-value]

        return self.t(seconds=td.total_seconds())


UUIDT = tp.TypeVar("UUIDT", bound=uuid.UUID)


class UUIDUnmarshaller(AbstractUnmarshaller[UUIDT], tp.Generic[UUIDT]):
    """Unmarshaller that converts an input to a [`uuid.UUID`][] (or subclasses).

    Note:
        The resolution algorithm is intentionally simple:

        1. Attempt to decode any bytes/string input into a real Python object.
        2. If the value is an integer, pass it into the constructor via the `int=` param.
        3. Otherwise, pass into the constructor directly.

    Tip:
        While the [`uuid.UUID`][] constructor supports many different keyword
        inputs for different types of UUID formats/encodings, we don't have a great
        method for detecting the correct input. We have moved with the assumption that
        the two most common formats are a standard string encoding, or an integer encoding.
    """

    def __call__(self, val: tp.Any) -> UUIDT:
        """Unmarshal a value into the bound `UUIDT` type.

        Args:
            val: The input value to unmarshal.

        See Also:
            - [`typelib.serdes.load`][]
        """
        decoded = serdes.load(val)
        if isinstance(decoded, int):
            return self.t(int=decoded)
        if isinstance(decoded, self.t):
            return decoded
        return self.t(decoded)  # type: ignore[arg-type]


PatternT = tp.TypeVar("PatternT", bound=re.Pattern)


class PatternUnmarshaller(AbstractUnmarshaller[PatternT], tp.Generic[PatternT]):
    """Unmarshaller that converts an input to a[`re.Pattern`][].

    Note:
        You can't instantiate a [`re.Pattern`][] directly, so we don't have a good
        method for handling patterns from a different library out-of-the-box. We simply call
        `re.compile()` on the decoded input.

    See Also:
        - [`typelib.serdes.decode`][]
    """

    def __call__(self, val: tp.Any) -> PatternT:
        decoded = serdes.decode(val)
        return re.compile(decoded)  # type: ignore[return-value]


class CastUnmarshaller(AbstractUnmarshaller[T]):
    """Unmarshaller that converts an input to an instance of `T` with a direct cast.

    Note:
        Before casting to the bound type, we will attempt to decode the value into a
        real Python object.

    See Also:
        - [`typelib.serdes.load`][]
    """

    __slots__ = ("caster",)

    def __init__(self, t: type[T], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context (unused).
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.caster: tp.Callable[[tp.Any], T] = self.origin  # type: ignore[assignment]

    def __call__(self, val: tp.Any) -> T:
        """Unmarshal a value into the bound `T` type.

        Args:
            val: The input value to unmarshal.
        """
        # Try to load the string, if this is JSON or a literal expression.
        decoded = serdes.load(val)
        # Short-circuit cast if we have the type we want.
        if isinstance(decoded, self.t):
            return decoded
        # Cast the decoded value to the type.
        return self.caster(decoded)


PathUnmarshaller = CastUnmarshaller[pathlib.Path]
MappingUnmarshaller = CastUnmarshaller[tp.Mapping]
IterableUnmarshaller = CastUnmarshaller[tp.Iterable]


LiteralT = tp.TypeVar("LiteralT")


class LiteralUnmarshaller(AbstractUnmarshaller[LiteralT], tp.Generic[LiteralT]):
    """Unmarshaller that will enforce an input conform to a defined [`typing.Literal`][].

    Note:
        We will attempt to decode the value into a real Python object if the input
        fails initial membership evaluation.

    See Also:
        - [`typelib.serdes.load`][]
    """

    __slots__ = ("values",)

    def __init__(self, t: type[LiteralT], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context (unused).
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.values = inspection.args(t)

    def __call__(self, val: tp.Any) -> LiteralT:
        if val in self.values:
            return val
        decoded = serdes.load(val)
        if decoded in self.values:
            return decoded  # type: ignore[return-value]

        raise ValueError(f"{decoded!r} is not one of {self.values!r}")


UnionT = tp.TypeVar("UnionT")


class UnionUnmarshaller(AbstractUnmarshaller[UnionT], tp.Generic[UnionT]):
    """Unmarshaller that will convert an input to one of the types defined in a [`typing.Union`][].

    Note:
        Union deserialization is messy and violates a static type-checking mechanism -
        for static type-checkers, `str | int` is equivalent to `int | str`. This breaks
        down during unmarshalling for the simple fact that casting something to `str`
        will always succeed, so we would never actually unmarshal the input it an `int`,
        even if that is the "correct" result.

        Our algorithm is intentionally simple:

        1. We iterate through each union member from top to bottom and call the
           resolved unmarshaller, returning the result.
        2. If any of `(ValueError, TypeError, SyntaxError)`, try again with the
           next unmarshaller.
        3. If all unmarshallers fail, then we have an invalid input, raise an error.

    Tip: TL;DR
        In order to ensure correctness, you should treat your union members as a stack,
        sorted from most-strict initialization to least-strict.
    """

    __slots__ = ("stack", "ordered_routines")

    def __init__(self, t: type[UnionT], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the member unmarshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.stack = inspection.args(t)
        self.ordered_routines = [self.context[typ] for typ in self.stack]

    def __call__(self, val: tp.Any) -> UnionT:
        """Unmarshal a value into the bound `UnionT`.

        Args:
            val: The input value to unmarshal.

        Raises:
            ValueError: If `val` cannot be unmarshalled into any member type.
        """
        for routine in self.ordered_routines:
            with contextlib.suppress(ValueError, TypeError, SyntaxError):
                unmarshalled = routine(val)
                return unmarshalled

        raise ValueError(f"{val!r} is not one of types {self.stack!r}")


_KT = tp.TypeVar("_KT")
_VT = tp.TypeVar("_VT")


MappingT = tp.TypeVar("MappingT", bound=tp.Mapping)


class SubscriptedMappingUnmarshaller(
    AbstractUnmarshaller[MappingT], tp.Generic[MappingT]
):
    """Unmarshaller for a subscripted mapping type.

    Note:
        This unmarshaller handles standard key->value mappings. We leverage our own
        generic `iteritems` to allow for translating other collections or structured
        objects into the target mapping.

        The algorithm is as follows:

        1. We attempt to decode the input into a real Python object.
        2. We iterate over key->value pairs.
        3. We call the key-type's unmarshaller on the `key` members.
        4. We call the value-type's unmarshaller on the `value` members.
        5. We pass the unmarshalling iterator in to the type's constructor.

    See Also:
        - [`typelib.serdes.load`][]
        - [`typelib.serdes.iteritems`][]
    """

    __slots__ = (
        "keys",
        "values",
    )

    def __init__(self, t: type[MappingT], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the member unmarshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        key_t, value_t = inspection.args(t)
        self.keys = context[key_t]
        self.values = context[value_t]

    def __call__(self, val: tp.Any) -> MappingT:
        """Unmarshal a value into the bound `MappingT`.

        Args:
            val: The input value to unmarshal.
        """
        # Always decode bytes.
        decoded = serdes.load(val)
        keys = self.keys
        values = self.values
        return self.origin(  # type: ignore[call-arg]
            ((keys(k), values(v)) for k, v in serdes.iteritems(decoded))
        )


IterableT = tp.TypeVar("IterableT", bound=tp.Iterable)


class SubscriptedIterableUnmarshaller(
    AbstractUnmarshaller[IterableT], tp.Generic[IterableT]
):
    """Unmarshaller for a subscripted iterable type.

    Note:
        This unmarshaller handles standard simple iterable types. We leverage our own
        generic `itervalues` to allow for translating other collections or structured
        objects into the target iterable.

        The algorithm is as follows:

        1. We attempt to decode the input into a real Python object.
        2. We iterate over values in the decoded input.
        3. We call the value-type's unmarshaller on the `value` members.
        5. We pass the unmarshalling iterator in to the type's constructor.

    See Also:
        - [`typelib.serdes.load`][]
        - [`typelib.serdes.itervalues`][]
    """

    __slots__ = ("values",)

    def __init__(
        self, t: type[IterableT], context: ContextT, *, var: str | None = None
    ):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the member unmarshaller.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t=t, context=context, var=var)
        # supporting tuple[str, ...]
        (value_t, *_) = inspection.args(t)
        self.values = context[value_t]

    def __call__(self, val: tp.Any) -> IterableT:
        """Unmarshal a value into the bound `IterableT`.

        Args:
            val: The input value to unmarshal.
        """
        # Always decode bytes.
        decoded = serdes.load(val)
        values = self.values
        return self.origin((values(v) for v in serdes.itervalues(decoded)))  # type: ignore[call-arg]


IteratorT = tp.TypeVar("IteratorT", bound=tp.Iterator)


class SubscriptedIteratorUnmarshaller(
    AbstractUnmarshaller[IteratorT], tp.Generic[IteratorT]
):
    """Unmarshaller for a subscripted iterator type.

    Note:
        This unmarshaller handles standard simple iterable types. We leverage our own
        generic `itervalues` to allow for translating other collections or structured
        objects into the target iterator.

        The algorithm is as follows:

        1. We attempt to decode the input into a real Python object.
        2. We iterate over values in the decoded input.
        3. We call the value-type's unmarshaller on the `value` members.
        5. We return a new, unmarshalling iterator.

    See Also:
        - [`typelib.serdes.load`][]
        - [`typelib.serdes.itervalues`][]
    """

    __slots__ = ("values",)

    def __init__(
        self, t: type[IteratorT], context: ContextT, *, var: str | None = None
    ):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the member unmarshaller.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        (value_t,) = inspection.args(t)
        self.values = context[value_t]

    def __call__(self, val: tp.Any) -> IteratorT:
        """Unmarshal a value into the bound `IteratorT`.

        Args:
            val: The input value to unmarshal.
        """
        # Always decode bytes.
        decoded = serdes.load(val)
        values = self.values
        it: IteratorT = (values(v) for v in serdes.itervalues(decoded))  # type: ignore[assignment]
        return it


class FixedTupleUnmarshaller(AbstractUnmarshaller[compat.TupleT]):
    """Unmarshaller for a "fixed" tuple (e.g., `tuple[int, str, float]`).

    Note:
        Python supports two distinct uses for tuples, unlike in other languages:

        1. Tuples with a fixed number of members.
        2. Tuples of variable length (an immutable sequence).

        "Fixed" tuples may have a distinct type for each member, while variable-length
        tuples may only have a single type (or union of types) for all members.

        Variable-length tuples are handled by our generic iterable unmarshaller.

        For "fixed" tuples, the algorithm is:

        1. Attempt to decode the input into a real Python object.
        2. zip the stack of member unmarshallers and the values in the decoded object.
        3. Unmarshal each value using the associated unmarshaller for that position.
        4. Pass the unmarshalling iterator in to the type's constructor.

    Tip:
        If the input has more members than the type definition allows, those members
        will be dropped by nature of our unmarshalling algorithm.

    See Also:
        - [`typelib.serdes.load`][]
        - [`typelib.serdes.itervalues`][]
    """

    __slots__ = ("ordered_routines", "stack")

    def __init__(
        self, t: type[compat.TupleT], context: ContextT, *, var: str | None = None
    ):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the value unmarshaller stack.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.stack = inspection.args(t)
        self.ordered_routines = [self.context[vt] for vt in self.stack]

    def __call__(self, val: tp.Any) -> compat.TupleT:
        """Unmarshal a value into the bound [`tuple`][] structure.

        Args:
            val: The input value to unmarshal.
        """
        decoded = serdes.load(val)
        return self.origin(
            routine(v)
            for routine, v in zip(self.ordered_routines, serdes.itervalues(decoded))
        )


_ST = tp.TypeVar("_ST")


class StructuredTypeUnmarshaller(AbstractUnmarshaller[_ST]):
    """Unmarshaller for a "structured" (user-defined) type.

    Note:
        This unmarshaller supports the unmarshalling of any mapping or structured
        type into the targeted structured type. There are limitations.

        The algorithm is:

        1. Attempt to decode the input into a real Python object.
        2. Using a mapping of the structured types "field" to the field-type's unmarshaller,
           iterate over the field->value pairs of the input, skipping fields in the
           input which are not present in the field mapping.
        3. Store each unmarshalled value in a keyword-argument mapping.
        4. Unpack the keyword argument mapping into the bound type's constructor.

    Tip:
        While we don't currently support arbitrary collections, we may add this
        functionality at a later date. Doing so requires more advanced introspection
        and parameter-binding that would lead to a significant loss in performance if
        not done carefully.

    See Also:
        - [`typelib.serdes.load`][]
        - [`typelib.serdes.itervalues`][]
    """

    __slots__ = ("fields_by_var",)

    def __init__(self, t: type[_ST], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the value field-to-unmarshaller mapping.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.fields_by_var = {m.var: m for m in self.context.values() if m.var}

    def __call__(self, val: tp.Any) -> _ST:
        """Unmarshal a value into the bound type.

        Args:
            val: The input value to unmarshal.
        """
        decoded = serdes.load(val)
        fields = self.fields_by_var
        kwargs = {f: fields[f](v) for f, v in serdes.iteritems(decoded) if f in fields}
        return self.t(**kwargs)



"""Type-specific logic for marshalling Python objects in to simple types for representation over-the-wire."""

from __future__ import annotations

import abc
import contextlib
import datetime
import decimal
import fractions
import pathlib
import re
import typing as tp
import uuid

from typelib import graph, serdes
from typelib.py import compat, inspection

T = tp.TypeVar("T")

__all__ = (
    "AbstractMarshaller",
    "ContextT",
    "BytesMarshaller",
    "StringMarshaller",
    "IntegerMarshaller",
    "FloatMarshaller",
    "DecimalMarshaller",
    "FractionMarshaller",
    "DateMarshaller",
    "DateTimeMarshaller",
    "TimeMarshaller",
    "TimeDeltaMarshaller",
    "UUIDMarshaller",
    "PathMarshaller",
    "PatternMarshaller",
    "MappingMarshaller",
    "IterableMarshaller",
    "LiteralMarshaller",
    "UnionMarshaller",
    "SubscriptedIterableMarshaller",
    "SubscriptedMappingMarshaller",
    "FixedTupleMarshaller",
    "StructuredTypeMarshaller",
)


class AbstractMarshaller(abc.ABC, tp.Generic[T]):
    """Abstract base class defining the common interface for marshallers.

    Marshallers are custom callables which maintain type-specific information. They
    use this information to provide robust, performant logic reducing Python objects
    into their primitive representations for over-the-wire encoding.

    Marshallers support contextual serialization, which enables the marshalling of
    nested types.

    Attributes:
        t: The root type of this marshaller.
        origin: If `t` is a generic, this will be an actionable runtime type
                related to `t`, otherwise it is the same as `t`.
        context: The complete type context for this unmarshaller.
        var: If this unmarshaller is used in a nested context, this will reference the
             field/parameter/index at which this unmarshaller should be used.
    """

    t: type[T]
    origin: type[T]
    context: ContextT
    var: str | None

    __slots__ = ("t", "origin", "context", "var")

    def __repr__(self):
        return f"<{self.__class__.__name__}(type={self.t!r}, origin={self.origin!r}, var={self.var!r})>"

    def __init__(self, t: type[T], context: ContextT, *, var: str | None = None):
        """Construct a marshaller instance.

        Args:
            t: The root type of this marshaller.
            context: The complete type context for this marshaller.
            var: The associated field or parameter name for this unmarshaller (optional).
        """
        self.t = t
        self.origin = inspection.origin(self.t)
        self.context = context
        self.var = var

    @abc.abstractmethod
    def __call__(self, val: T) -> serdes.MarshalledValueT: ...


ContextT: tp.TypeAlias = "tp.Mapping[type | graph.TypeNode, AbstractMarshaller]"


class NoOpMarshaller(AbstractMarshaller[T], tp.Generic[T]):
    """A marshaller that does nothing."""

    def __call__(self, val: T) -> serdes.MarshalledValueT:
        """Run the marshaller.

        Args:
            val: The value to marshal.
        """
        return val  # type: ignore[return-value]


BytesMarshaller = NoOpMarshaller[bytes]


class CastMarshaller(AbstractMarshaller[T], tp.Generic[T]):
    """A marshaller that casts a value to a specific type."""

    def __call__(self, val: T) -> serdes.MarshalledValueT:
        """Marshal a the value into bound type.

        Args:
            val: The value to marshal.
        """
        cast = tp.cast("serdes.MarshalledValueT", self.origin(val))  # type: ignore[call-arg]
        return cast


IntegerMarshaller = CastMarshaller[int]
FloatMarshaller = CastMarshaller[float]


class ToStringMarshaller(AbstractMarshaller[T], tp.Generic[T]):
    """A marshaller that converts a value to a string."""

    def __call__(self, val: T) -> str:
        """Marshal a value of the bound type into a string.

        Args:
            val: The value to marshal.
        """
        return str(val)


StringMarshaller = ToStringMarshaller[str]

DecimalT = tp.TypeVar("DecimalT", bound=decimal.Decimal)
DecimalMarshaller = ToStringMarshaller[DecimalT]

FractionT = tp.TypeVar("FractionT", bound=fractions.Fraction)
FractionMarshaller = ToStringMarshaller[FractionT]

UUIDT = tp.TypeVar("UUIDT", bound=uuid.UUID)
UUIDMarshaller = ToStringMarshaller[UUIDT]

PathT = tp.TypeVar("PathT", bound=pathlib.Path)
PathMarshaller = ToStringMarshaller[PathT]


PatternT = tp.TypeVar("PatternT", bound=re.Pattern)


class PatternMarshaller(AbstractMarshaller[PatternT]):
    """A marshaller that converts a [`re.Pattern`][] to a string."""

    def __call__(self, val: PatternT) -> str:
        """Marshal a compiled regex pattern into a string.

        Args:
            val: The pattern to marshal.
        """
        return val.pattern


DateOrTimeT = tp.TypeVar(
    "DateOrTimeT", datetime.date, datetime.time, datetime.timedelta
)


class ToISOTimeMarshaller(AbstractMarshaller[DateOrTimeT], tp.Generic[DateOrTimeT]):
    """A marshaller that converts any date/time object to a ISO time string.

    See Also:
        - [`typelib.serdes.isoformat`]
    """

    def __call__(self, val: DateOrTimeT) -> str:
        """Marshal a date/time object into a ISO time string.

        Args:
            val: The date/time object to marshal.
        """
        return serdes.isoformat(val)


DateT = tp.TypeVar("DateT", bound=datetime.date)
DateMarshaller = ToISOTimeMarshaller[DateT]

DateTimeT = tp.TypeVar("DateTimeT", bound=datetime.datetime)
DateTimeMarshaller = ToISOTimeMarshaller[DateTimeT]

TimeT = tp.TypeVar("TimeT", bound=datetime.time)
TimeMarshaller = ToISOTimeMarshaller[TimeT]

TimeDeltaT = tp.TypeVar("TimeDeltaT", bound=datetime.timedelta)
TimeDeltaMarshaller = ToISOTimeMarshaller[TimeDeltaT]


LiteralT = tp.TypeVar("LiteralT")


class LiteralMarshaller(AbstractMarshaller[LiteralT], tp.Generic[LiteralT]):
    """A marshaller that enforces the given value be one of the values in the defined [`typing.Literal`][]"""

    __slots__ = ("values",)

    def __init__(self, t: type[LiteralT], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The Literal type to enforce.
            context: Nested type context (unused).
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.values = inspection.args(t)

    def __call__(self, val: LiteralT) -> serdes.MarshalledValueT:
        """Enforce the given value is a member of the bound `Literal` type.

        Args:
            val: The value to enforce.

        Raises:
            ValueError: If `val` is not a member of the bound `Literal` type.
        """
        if val in self.values:
            return val  # type: ignore[return-value]

        raise ValueError(f"{val!r} is not one of {self.values!r}")


UnionT = tp.TypeVar("UnionT")


class UnionMarshaller(AbstractMarshaller[UnionT], tp.Generic[UnionT]):
    """A marshaller for dumping a given value via one of the types in the defined bound union.

    See Also:
        - [`UnionUnmarshaller`][typelib.unmarshals.routines.UnionUnmarshaller]
    """

    __slots__ = ("stack", "ordered_routines")

    def __init__(self, t: type[UnionT], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshal into.
            context: Any nested type context. Used to resolve the member marshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.stack = inspection.args(t)
        self.ordered_routines = [self.context[typ] for typ in self.stack]

    def __call__(self, val: UnionT) -> serdes.MarshalledValueT:
        """Unmarshal a value into the bound `UnionT`.

        Args:
            val: The input value to unmarshal.

        Raises:
            ValueError: If `val` cannot be marshalled via any member type.
        """
        for routine in self.ordered_routines:
            with contextlib.suppress(ValueError, TypeError, SyntaxError):
                unmarshalled = routine(val)
                return unmarshalled

        raise ValueError(f"{val!r} is not one of types {self.stack!r}")


MappingT = tp.TypeVar("MappingT", bound=tp.Mapping)


class MappingMarshaller(AbstractMarshaller[MappingT], tp.Generic[MappingT]):
    """A marshaller for dumping any mapping into a simple [`dict`][]."""

    def __call__(self, val: MappingT) -> MarshalledMappingT:
        """Marshal a mapping into a simple [`dict`][].

        Args:
            val: The mapping object to marshal.
        """
        return {**val}


IterableT = tp.TypeVar("IterableT", bound=tp.Iterable)


class IterableMarshaller(AbstractMarshaller[IterableT], tp.Generic[IterableT]):
    """A marshaller for dumping any iterable into a simple [`list`][]."""

    def __call__(self, val: IterableT) -> MarshalledIterableT:
        """Marshal an iterable into a simple [`list`][].

        Args:
            val: The iterable to marshal.
        """
        return [*val]


class SubscriptedMappingMarshaller(AbstractMarshaller[MappingT], tp.Generic[MappingT]):
    """A marshaller for dumping a subscripted mapping into a simple [`dict`][].

    Keys are marshalled according to the defined key-type, values according to the defined value-type.

    See Also:
        - [`SubscriptedMappingUnmarshaller`][typelib.unmarshals.routines.SubscriptedMappingUnmarshaller]
    """

    __slots__ = (
        "keys",
        "values",
    )

    def __init__(self, t: type[MappingT], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshals from.
            context: Any nested type context. Used to resolve the member marshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        key_t, value_t = inspection.args(t)
        self.keys = context[key_t]
        self.values = context[value_t]

    def __call__(self, val: MappingT) -> MarshalledMappingT:
        keys = self.keys
        values = self.values
        return {keys(k): values(v) for k, v in serdes.iteritems(val)}  # type: ignore[misc]


class SubscriptedIterableMarshaller(
    AbstractMarshaller[IterableT], tp.Generic[IterableT]
):
    """A marshaller for dumping a subscripted iterable into a simple [`list`][].

    Values are marshalled according to the defined value-type.

    See Also:
        - [`SubscriptedIterableUnmarshaller`][typelib.unmarshals.routines.SubscriptedIterableUnmarshaller]
    """

    __slots__ = ("values",)

    def __init__(
        self, t: type[IterableT], context: ContextT, *, var: str | None = None
    ):
        """Constructor.

        Args:
            t: The type to unmarshals from.
            context: Any nested type context. Used to resolve the member marshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t=t, context=context, var=var)
        # supporting tuple[str, ...]
        (value_t, *_) = inspection.args(t)
        self.values = context[value_t]

    def __call__(self, val: IterableT) -> MarshalledIterableT:
        """Marshal an iterable into a simple [`list`][].

        Args:
            val: The iterable to marshal.
        """
        # Always decode bytes.
        values = self.values
        return [values(v) for v in serdes.itervalues(val)]


class FixedTupleMarshaller(AbstractMarshaller[compat.TupleT]):
    """A marshaller for dumping a "fixed" tuple to a simple [`list`][].

    Values are marshalled according to the value-type in the order they are defined.

    See Also:
        - [`FixedTupleUnmarshaller`][typelib.unmarshals.routines.FixedTupleUnmarshaller]
    """

    __slots__ = ("ordered_routines", "stack")

    def __init__(
        self, t: type[compat.TupleT], context: ContextT, *, var: str | None = None
    ):
        """Constructor.

        Args:
            t: The type to unmarshal from.
            context: Any nested type context. Used to resolve the member marshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.stack = inspection.args(t)
        self.ordered_routines = [self.context[vt] for vt in self.stack]

    def __call__(self, val: compat.TupleT) -> MarshalledIterableT:
        """Marshal a tuple into a simple [`list`][].

        Args:
            val: The tuple to marshal.
        """
        return [
            routine(v)
            for routine, v in zip(self.ordered_routines, serdes.itervalues(val))
        ]


_ST = tp.TypeVar("_ST")


class StructuredTypeMarshaller(AbstractMarshaller[_ST]):
    """A marshaller for dumping a structured (user-defined) type to a simple [`dict`][].

    See Also:
        - [`StructuredTypeUnmarshaller`][typelib.unmarshals.routines.StructuredTypeUnmarshaller]
    """

    __slots__ = ("fields_by_var",)

    def __init__(self, t: type[_ST], context: ContextT, *, var: str | None = None):
        """Constructor.

        Args:
            t: The type to unmarshals from.
            context: Any nested type context. Used to resolve the member marshallers.
            var: A variable name for the indicated type annotation (unused, optional).
        """
        super().__init__(t, context, var=var)
        self.fields_by_var = {m.var: m for m in self.context.values() if m.var}

    def __call__(self, val: _ST) -> MarshalledMappingT:
        """Marshal a structured type into a simple [`dict`][].

        Args:
            val: The structured type to marshal.
        """
        fields = self.fields_by_var
        return {f: fields[f](v) for f, v in serdes.iteritems(val) if f in fields}


MarshalledMappingT: tp.TypeAlias = dict[
    serdes.PythonPrimitiveT, serdes.MarshalledValueT
]
MarshalledIterableT: tp.TypeAlias = list[serdes.MarshalledValueT]
